"""Search algorithms for Transducer models."""

from typing import (
    Any,
    Dict,
    List,
    Optional,
    Tuple,
    Union
)

import numpy as np
import onnxruntime
from scipy.special import log_softmax

from espnet_onnx.utils.function import is_prefix
from espnet_onnx.utils.function import recombine_hyps
from espnet_onnx.utils.function import select_k_expansions
from espnet_onnx.utils.function import subtract
from espnet_onnx.utils.function import topk
from espnet_onnx.asr.beam_search.hyps import TransducerHypothesis
from espnet_onnx.asr.beam_search.hyps import ExtendedHypothesis


class BeamSearchTransducer:
    """Beam search implementation for Transducer."""

    def __init__(
        self,
        bs_config,
        token_config,
        scorers: List[Any],
        weights: Dict[str, float],
    ):
        """Initialize Transducer search module.
        Args:
            decoder: Decoder module.
            joint_network: Joint network module.
            beam_size: Beam size.
            lm: LM class.
            lm_weight: LM weight for soft fusion.
            search_type: Search algorithm to use during inference.
            max_sym_exp: Number of maximum symbol expansions at each time step. (TSD)
            u_max: Maximum output sequence length. (ALSD)
            nstep: Number of maximum expansion steps at each time step. (NSC/mAES)
            prefix_alpha: Maximum prefix length in prefix search. (NSC/mAES)
            expansion_beta:
              Number of additional candidates for expanded hypotheses selection. (mAES)
            expansion_gamma: Allowed logp difference for prune-by-value method. (mAES)
            score_norm: Normalize final scores by length. ("default")
            nbest: Number of final hypothesis.
        """
        self.decoder = scorers['decoder']
        self.joint_network = scorers['joint_network']

        self.beam_size = bs_config.beam_size
        self.vocab_size = len(token_config.list)
        self.blank_id = token_config.blank

        if self.beam_size <= 1:
            self.search_algorithm = self.greedy_search

        elif bs_config.search_type == "default":
            self.search_algorithm = self.default_beam_search

        elif bs_config.search_type == "tsd":
            self.max_sym_exp = bs_config.search_args.max_sym_exp
            self.search_algorithm = self.time_sync_decoding

        elif bs_config.search_type == "alsd":
            self.u_max = bs_config.search_args.u_max
            self.search_algorithm = self.align_length_sync_decoding

        elif bs_config.search_type == "nsc":
            self.nstep = bs_config.search_args.nstep
            self.prefix_alpha = bs_config.search_args.prefix_alpha
            self.search_algorithm = self.nsc_beam_search

        elif bs_config.search_type == "maes":
            self.nstep = bs_config.search_args.nstep if bs_config.search_args.nstep > 1 else 2
            self.prefix_alpha = bs_config.search_args.prefix_alpha
            self.expansion_gamma = bs_config.search_args.expansion_gamma
            self.expansion_beta = bs_config.search_args.expansion_beta
            self.search_algorithm = self.modified_adaptive_expansion_search

        else:
            raise NotImplementedError

        self.use_lm = 'lm' in scorers.keys()
        if self.use_lm:
            self.lm = scorers['lm']
            self.lm_weight = weights['lm']
        else:
            self.lm = None
            self.lm_weight = None

        self.score_norm = bs_config.score_norm

    def __call__(
        self, enc_out: np.ndarray
    ) -> Union[List[TransducerHypothesis], List[ExtendedHypothesis]]:
        """Perform beam search.

        Args:
            enc_out: Encoder output sequence. (T, D_enc)

        Returns:
            nbest_hyps: N-best decoding results

        """
        nbest_hyps = self.search_algorithm(enc_out)
        return nbest_hyps

    def sort_nbest(
        self, hyps: Union[List[TransducerHypothesis], List[ExtendedHypothesis]]
    ) -> Union[List[TransducerHypothesis], List[ExtendedHypothesis]]:
        """Sort hypotheses by score or score given sequence length.

        Args:
            hyps: TransducerHypothesis.

        Return:
            hyps: Sorted hypothesis.

        """
        if self.score_norm:
            hyps.sort(key=lambda x: x.score / len(x.yseq), reverse=True)
        else:
            hyps.sort(key=lambda x: x.score, reverse=True)

        return hyps

    def prefix_search(
        self, hyps: List[ExtendedHypothesis], enc_out_t: np.ndarray
    ) -> List[ExtendedHypothesis]:
        """Prefix search for NSC and mAES strategies.
        Based on https://arxiv.org/pdf/1211.3711.pdf
        """
        for j, hyp_j in enumerate(hyps[:-1]):
            for hyp_i in hyps[(j + 1):]:
                curr_id = len(hyp_j.yseq)
                pref_id = len(hyp_i.yseq)

                if (
                    is_prefix(hyp_j.yseq, hyp_i.yseq)
                    and (curr_id - pref_id) <= self.prefix_alpha
                ):
                    logp = log_softmax(self.joint_network(
                        enc_out_t[None, :], hyp_i.dec_out[-1][None, :]), axis=-1)[0]
                    curr_score = hyp_i.score + float(logp[hyp_j.yseq[pref_id]])

                    for k in range(pref_id, (curr_id - 1)):
                        logp = log_softmax(self.joint_network(
                            enc_out_t[None, :], hyp_i.dec_out[-1][None, :]), axis=-1)[0]
                        curr_score += float(logp[hyp_j.yseq[k + 1]])

                    hyp_j.score = np.logaddexp(hyp_j.score, curr_score)

        return hyps

    def greedy_search(self, enc_out: np.ndarray) -> List[TransducerHypothesis]:
        """Greedy search implementation.

        Args:
            enc_out: Encoder output sequence. (T, D_enc)

        Returns:
            hyp: 1-best hypotheses.

        """
        dec_state = self.decoder.init_state()

        hyp = TransducerHypothesis(
            score=0.0, yseq=[self.blank_id], dec_state=dec_state)
        cache = {}

        dec_out, state, _ = self.decoder.score(hyp, cache)

        for enc_out_t in enc_out:
            logp = log_softmax(self.joint_network(enc_out_t, dec_out), axis=-1)
            top_logp = np.max(logp, axis=-1)
            pred = np.argmax(logp, axis=-1)

            if pred != self.blank_id:
                hyp.yseq.append(int(pred))
                hyp.score += float(top_logp)

                hyp.dec_state = state
                dec_out, state, _ = self.decoder.score(hyp, cache)

        return [hyp]

    def default_beam_search(self, enc_out: np.ndarray) -> List[TransducerHypothesis]:
        """Beam search implementation.
        Modified from https://arxiv.org/pdf/1211.3711.pdf

        Args:
            enc_out: Encoder output sequence. (T, D)

        Returns:
            nbest_hyps: N-best hypothesis.

        """
        beam = min(self.beam_size, self.vocab_size)
        beam_k = min(beam, (self.vocab_size - 1))

        dec_state = self.decoder.init_state()

        kept_hyps = [TransducerHypothesis(
            score=0.0, yseq=[self.blank_id], dec_state=dec_state)]
        cache = {}

        for enc_out_t in enc_out:
            hyps = kept_hyps
            kept_hyps = []

            while True:
                max_hyp = max(hyps, key=lambda x: x.score)
                hyps.remove(max_hyp)

                dec_out, state, lm_tokens = self.decoder.score(max_hyp, cache)

                logp = log_softmax(self.joint_network(
                    enc_out_t, dec_out), axis=-1)
                top_k = topk(logp[1:], beam_k, require_value=True)
                kept_hyps.append(
                    TransducerHypothesis(
                        score=(max_hyp.score + float(logp[0:1])),
                        yseq=max_hyp.yseq[:],
                        dec_state=max_hyp.dec_state,
                        lm_state=max_hyp.lm_state,
                    )
                )

                # NOTE: we do not have score function in lm
                if self.use_lm:
                    lm_scores, lm_state = self.lm.score(
                        lm_tokens, max_hyp.lm_state, None
                    )
                else:
                    lm_state = max_hyp.lm_state

                for logp, k in zip(*top_k):
                    score = max_hyp.score + float(logp)

                    if self.use_lm:
                        score += self.lm_weight * lm_scores[k + 1]

                    hyps.append(
                        TransducerHypothesis(
                            score=score,
                            yseq=max_hyp.yseq[:] + [int(k + 1)],
                            dec_state=state,
                            lm_state=lm_state,
                        )
                    )

                hyps_max = float(max(hyps, key=lambda x: x.score).score)
                kept_most_prob = sorted(
                    [hyp for hyp in kept_hyps if hyp.score > hyps_max],
                    key=lambda x: x.score,
                )
                if len(kept_most_prob) >= beam:
                    kept_hyps = kept_most_prob
                    break

        return self.sort_nbest(kept_hyps)

    def time_sync_decoding(self, enc_out: np.ndarray) -> List[TransducerHypothesis]:
        """Time synchronous beam search implementation.
        Based on https://ieeexplore.ieee.org/document/9053040

        Args:
            enc_out: Encoder output sequence. (T, D)

        Returns:
            nbest_hyps: N-best hypothesis.

        """
        beam = min(self.beam_size, self.vocab_size)
        beam_state = self.decoder.init_state(beam)

        B = [
            TransducerHypothesis(
                yseq=[self.blank_id],
                score=0.0,
                dec_state=self.decoder.select_state(beam_state, 0),
            )
        ]
        cache = {}

        if self.use_lm:
            B[0].lm_state = self.lm.zero_state()

        for enc_out_t in enc_out:
            A = []
            C = B

            enc_out_t = enc_out_t[None, :]

            for v in range(self.max_sym_exp):
                D = []
                beam_dec_out, beam_state, beam_lm_tokens = self.decoder.batch_score(
                    C,
                    beam_state,
                    cache,
                    self.use_lm,
                )
                beam_logp = log_softmax(self.joint_network(
                    enc_out_t, beam_dec_out), axis=-1)
                beam_topk = topk(beam_logp[:, 1:], beam, require_value=True)

                seq_A = [h.yseq for h in A]

                for i, hyp in enumerate(C):
                    if hyp.yseq not in seq_A:
                        A.append(
                            TransducerHypothesis(
                                score=(hyp.score + float(beam_logp[i, 0])),
                                yseq=hyp.yseq[:],
                                dec_state=hyp.dec_state,
                                lm_state=hyp.lm_state,
                            )
                        )
                    else:
                        dict_pos = seq_A.index(hyp.yseq)

                        A[dict_pos].score = np.logaddexp(
                            A[dict_pos].score, (hyp.score +
                                                float(beam_logp[i, 0]))
                        )

                if v < (self.max_sym_exp - 1):
                    if self.use_lm:
                        beam_lm_scores, beam_lm_states = self.lm.batch_score(
                            beam_lm_tokens, [c.lm_state for c in C], None
                        )

                    for i, hyp in enumerate(C):
                        for logp, k in zip(beam_topk[0][i], beam_topk[1][i] + 1):
                            new_hyp = TransducerHypothesis(
                                score=(hyp.score + float(logp)),
                                yseq=(hyp.yseq + [int(k)]),
                                dec_state=self.decoder.select_state(
                                    beam_state, i),
                                lm_state=hyp.lm_state,
                            )

                            if self.use_lm:
                                new_hyp.score += self.lm_weight * \
                                    beam_lm_scores[i, k]
                                new_hyp.lm_state = beam_lm_states[i]

                            D.append(new_hyp)

                C = sorted(D, key=lambda x: x.score, reverse=True)[:beam]

            B = sorted(A, key=lambda x: x.score, reverse=True)[:beam]

        return self.sort_nbest(B)

    def align_length_sync_decoding(self, enc_out: np.ndarray) -> List[TransducerHypothesis]:
        """Alignment-length synchronous beam search implementation.
        Based on https://ieeexplore.ieee.org/document/9053040

        Args:
            h: Encoder output sequences. (T, D)

        Returns:
            nbest_hyps: N-best hypothesis.

        """
        beam = min(self.beam_size, self.vocab_size)
        t_max = int(enc_out.shape[0])
        u_max = min(self.u_max, (t_max - 1))

        beam_state = self.decoder.init_state(beam)

        B = [
            TransducerHypothesis(
                yseq=[self.blank_id],
                score=0.0,
                dec_state=self.decoder.select_state(beam_state, 0),
            )
        ]
        final = []
        cache = {}

        if self.use_lm:
            B[0].lm_state = self.lm.zero_state()

        for i in range(t_max + u_max):
            A = []

            B_ = []
            B_enc_out = []
            for hyp in B:
                u = len(hyp.yseq) - 1
                t = i - u

                if t > (t_max - 1):
                    continue

                B_.append(hyp)
                B_enc_out.append((t, enc_out[t]))

            if B_:
                beam_dec_out, beam_state, beam_lm_tokens = self.decoder.batch_score(
                    B_,
                    beam_state,
                    cache,
                    self.use_lm,
                )

                beam_enc_out = np.stack([x[1] for x in B_enc_out])

                beam_logp = log_softmax(self.joint_network(
                    beam_enc_out, beam_dec_out), axis=-1)
                beam_topk = topk(beam_logp[:, 1:], beam, require_value=True)

                if self.use_lm:
                    beam_lm_scores, beam_lm_states = self.lm.batch_score(
                        beam_lm_tokens,
                        [b.lm_state for b in B_],
                        None,
                    )

                for i, hyp in enumerate(B_):
                    new_hyp = TransducerHypothesis(
                        score=(hyp.score + float(beam_logp[i, 0])),
                        yseq=hyp.yseq[:],
                        dec_state=hyp.dec_state,
                        lm_state=hyp.lm_state,
                    )

                    A.append(new_hyp)

                    if B_enc_out[i][0] == (t_max - 1):
                        final.append(new_hyp)

                    for logp, k in zip(beam_topk[0][i], beam_topk[1][i] + 1):
                        new_hyp = TransducerHypothesis(
                            score=(hyp.score + float(logp)),
                            yseq=(hyp.yseq[:] + [int(k)]),
                            dec_state=self.decoder.select_state(beam_state, i),
                            lm_state=hyp.lm_state,
                        )

                        if self.use_lm:
                            new_hyp.score += self.lm_weight * \
                                beam_lm_scores[i, k]
                            new_hyp.lm_state = beam_lm_states[i]

                        A.append(new_hyp)

                B = sorted(A, key=lambda x: x.score, reverse=True)[:beam]
                B = recombine_hyps(B)

        if final:
            return self.sort_nbest(final)
        else:
            return B

    def nsc_beam_search(self, enc_out: np.ndarray) -> List[ExtendedHypothesis]:
        """N-step constrained beam search implementation.
        Based on/Modified from https://arxiv.org/pdf/2002.03577.pdf.
        Please reference ESPnet (b-flo, PR #2444) for any usage outside ESPnet
        until further modifications.

        Args:
            enc_out: Encoder output sequence. (T, D_enc)

        Returns:
            nbest_hyps: N-best hypothesis.

        """
        beam = min(self.beam_size, self.vocab_size)
        beam_k = min(beam, (self.vocab_size - 1))
        beam_state = self.decoder.init_state(beam)

        init_tokens = [
            ExtendedHypothesis(
                yseq=[self.blank_id],
                score=0.0,
                dec_state=self.decoder.select_state(beam_state, 0),
            )
        ]

        cache = {}

        beam_dec_out, beam_state, beam_lm_tokens = self.decoder.batch_score(
            init_tokens,
            beam_state,
            cache,
            self.use_lm,
        )

        state = self.decoder.select_state(beam_state, 0)

        if self.use_lm:
            beam_lm_scores, beam_lm_states = self.lm.batch_score(
                beam_lm_tokens,
                [i.lm_state for i in init_tokens],
                None,
            )
            lm_state = beam_lm_states[0]
            lm_scores = beam_lm_scores[0]
        else:
            lm_state = None
            lm_scores = None

        kept_hyps = [
            ExtendedHypothesis(
                yseq=[self.blank_id],
                score=0.0,
                dec_state=state,
                dec_out=[beam_dec_out[0]],
                lm_state=lm_state,
                lm_scores=lm_scores,
            )
        ]

        for enc_out_t in enc_out:
            hyps = self.prefix_search(
                sorted(kept_hyps, key=lambda x: len(x.yseq), reverse=True),
                enc_out_t,
            )
            kept_hyps = []
            beam_enc_out = np.expand_dims(enc_out_t, 0)

            S = []
            V = []
            for n in range(self.nstep):
                beam_dec_out = np.stack([hyp.dec_out[-1] for hyp in hyps])

                beam_logp = log_softmax(self.joint_network(
                    beam_enc_out, beam_dec_out), axis=-1)
                beam_topk = topk(beam_logp[:, 1:], beam_k, require_value=True)

                for i, hyp in enumerate(hyps):
                    S.append(
                        ExtendedHypothesis(
                            yseq=hyp.yseq[:],
                            score=hyp.score + float(beam_logp[i, 0:1]),
                            dec_out=hyp.dec_out[:],
                            dec_state=hyp.dec_state,
                            lm_state=hyp.lm_state,
                            lm_scores=hyp.lm_scores,
                        )
                    )

                    for logp, k in zip(beam_topk[0][i], beam_topk[1][i] + 1):
                        score = hyp.score + float(logp)

                        if self.use_lm:
                            score += self.lm_weight * float(hyp.lm_scores[k])

                        V.append(
                            ExtendedHypothesis(
                                yseq=hyp.yseq[:] + [int(k)],
                                score=score,
                                dec_out=hyp.dec_out[:],
                                dec_state=hyp.dec_state,
                                lm_state=hyp.lm_state,
                                lm_scores=hyp.lm_scores,
                            )
                        )

                V.sort(key=lambda x: x.score, reverse=True)
                V = subtract(V, hyps)[:beam]

                beam_state = self.decoder.create_batch_states(
                    beam_state,
                    [v.dec_state for v in V],
                    [v.yseq for v in V],
                )
                beam_dec_out, beam_state, beam_lm_tokens = self.decoder.batch_score(
                    V,
                    beam_state,
                    cache,
                    self.use_lm,
                )

                if self.use_lm:
                    beam_lm_scores, beam_lm_states = self.lm.batch_score(
                        beam_lm_tokens, [v.lm_state for v in V], None
                    )

                if n < (self.nstep - 1):
                    for i, v in enumerate(V):
                        v.dec_out.append(beam_dec_out[i])

                        v.dec_state = self.decoder.select_state(beam_state, i)

                        if self.use_lm:
                            v.lm_state = beam_lm_states[i]
                            v.lm_scores = beam_lm_scores[i]

                    hyps = V[:]
                else:
                    beam_logp = log_softmax(self.joint_network(
                        beam_enc_out, beam_dec_out), axis=-1)

                    for i, v in enumerate(V):
                        if self.nstep != 1:
                            v.score += float(beam_logp[i, 0])

                        v.dec_out.append(beam_dec_out[i])
                        v.dec_state = self.decoder.select_state(beam_state, i)

                        if self.use_lm:
                            v.lm_state = beam_lm_states[i]
                            v.lm_scores = beam_lm_scores[i]

            kept_hyps = sorted(
                (S + V), key=lambda x: x.score, reverse=True)[:beam]

        return self.sort_nbest(kept_hyps)

    def modified_adaptive_expansion_search(
        self, enc_out: np.ndarray
    ) -> List[ExtendedHypothesis]:
        """It's the modified Adaptive Expansion Search (mAES) implementation.
        Based on/modified from https://ieeexplore.ieee.org/document/9250505 and NSC.

        Args:
            enc_out: Encoder output sequence. (T, D_enc)

        Returns:
            nbest_hyps: N-best hypothesis.

        """
        beam = min(self.beam_size, self.vocab_size)
        beam_state = self.decoder.init_state(beam)

        init_tokens = [
            ExtendedHypothesis(
                yseq=[self.blank_id],
                score=0.0,
                dec_state=self.decoder.select_state(beam_state, 0),
            )
        ]

        cache = {}

        beam_dec_out, beam_state, beam_lm_tokens = self.decoder.batch_score(
            init_tokens,
            beam_state,
            cache,
            self.use_lm,
        )

        state = self.decoder.select_state(beam_state, 0)

        if self.use_lm:
            beam_lm_scores, beam_lm_states = self.lm.batch_score(
                beam_lm_tokens, [i.lm_state for i in init_tokens], None
            )

            lm_state = beam_lm_states[0]
            lm_scores = beam_lm_scores[0]
        else:
            lm_state = None
            lm_scores = None

        kept_hyps = [
            ExtendedHypothesis(
                yseq=[self.blank_id],
                score=0.0,
                dec_state=state,
                dec_out=[beam_dec_out[0]],
                lm_state=lm_state,
                lm_scores=lm_scores,
            )
        ]

        for enc_out_t in enc_out:
            hyps = self.prefix_search(
                sorted(kept_hyps, key=lambda x: len(x.yseq), reverse=True),
                enc_out_t,
            )
            kept_hyps = []
            beam_enc_out = np.expand_dims(enc_out_t, 0)

            list_b = []
            for n in range(self.nstep):
                beam_dec_out = np.array([h.dec_out[-1] for h in hyps])

                beam_logp = log_softmax(self.joint_network(
                    beam_enc_out, beam_dec_out), axis=-1)
                k_expansions = select_k_expansions(
                    hyps, beam_logp, beam, self.expansion_gamma, self.expansion_beta
                )

                list_exp = []
                for i, hyp in enumerate(hyps):
                    for k, new_score in k_expansions[i]:
                        new_hyp = ExtendedHypothesis(
                            yseq=hyp.yseq[:],
                            score=new_score,
                            dec_out=hyp.dec_out[:],
                            dec_state=hyp.dec_state,
                            lm_state=hyp.lm_state,
                            lm_scores=hyp.lm_scores,
                        )

                        if k == 0:
                            list_b.append(new_hyp)
                        else:
                            new_hyp.yseq.append(int(k))

                            if self.use_lm:
                                new_hyp.score += self.lm_weight * float(
                                    hyp.lm_scores[k]
                                )

                            list_exp.append(new_hyp)

                if not list_exp:
                    kept_hyps = sorted(list_b, key=lambda x: x.score, reverse=True)[
                        :beam
                    ]

                    break
                else:
                    beam_state = self.decoder.create_batch_states(
                        beam_state,
                        [hyp.dec_state for hyp in list_exp],
                        [hyp.yseq for hyp in list_exp],
                    )

                    beam_dec_out, beam_state, beam_lm_tokens = self.decoder.batch_score(
                        list_exp,
                        beam_state,
                        cache,
                        self.use_lm,
                    )

                    if self.use_lm:
                        beam_lm_scores, beam_lm_states = self.lm.batch_score(
                            beam_lm_tokens, [
                                k.lm_state for k in list_exp], None
                        )

                    if n < (self.nstep - 1):
                        for i, hyp in enumerate(list_exp):
                            hyp.dec_out.append(beam_dec_out[i])
                            hyp.dec_state = self.decoder.select_state(
                                beam_state, i)

                            if self.use_lm:
                                hyp.lm_state = beam_lm_states[i]
                                hyp.lm_scores = beam_lm_scores[i]

                        hyps = list_exp[:]
                    else:
                        beam_logp = log_softmax(self.joint_network(
                            beam_enc_out, beam_dec_out), axis=-1)

                        for i, hyp in enumerate(list_exp):
                            hyp.score += float(beam_logp[i, 0])

                            hyp.dec_out.append(beam_dec_out[i])
                            hyp.dec_state = self.decoder.select_state(
                                beam_state, i)

                            if self.use_lm:
                                hyp.lm_states = beam_lm_states[i]
                                hyp.lm_scores = beam_lm_scores[i]

                        kept_hyps = sorted(
                            list_b + list_exp, key=lambda x: x.score, reverse=True
                        )[:beam]

        return self.sort_nbest(kept_hyps)



from typing import (
    Tuple,
    List
)

import onnxruntime
import numpy as np

from espnet_onnx.asr.frontend.frontend import Frontend
from espnet_onnx.asr.frontend.global_mvn import GlobalMVN
from espnet_onnx.asr.frontend.utterance_mvn import UtteranceMVN
from espnet_onnx.utils.function import (
    make_pad_mask,
    mask_fill
)
from espnet_onnx.utils.config import Config


class Encoder:
    def __init__(
        self,
        encoder_config: Config,
        providers: List[str],
        use_quantized: bool = False,
    ):
        self.config = encoder_config
        if use_quantized:
            self.encoder = onnxruntime.InferenceSession(
                self.config.quantized_model_path,
                providers=providers
            )
        else:
            self.encoder = onnxruntime.InferenceSession(
                self.config.model_path,
                providers=providers
            )

        self.frontend = Frontend(self.config.frontend)
        if self.config.do_normalize:
            if self.config.normalize.type == 'gmvn':
                self.normalize = GlobalMVN(self.config.normalize)
            elif self.config.normalize.type == 'utterance_mvn':
                self.normalize = UtteranceMVN(self.config.normalize)

        # if self.config.do_preencoder:
        #     self.preencoder = Preencoder(self.config.preencoder)

        # if self.config.do_postencoder:
        #     self.postencoder = Postencoder(self.config.postencoder)

    def __call__(
        self, speech: np.ndarray, speech_length: np.ndarray
    ) -> Tuple[np.ndarray, np.ndarray]:
        """Frontend + Encoder. Note that this method is used by asr_inference.py
        Args:
            speech: (Batch, Length, ...)
            speech_lengths: (Batch, )
        """
        # 1. Extract feature
        feats, feat_length = self.frontend(speech, speech_length)

        # 2. normalize with global MVN
        if self.config.do_normalize:
            feats, feat_length = self.normalize(feats, feat_length)

        # if self.config.do_preencoder:
        #     feats, feats_lengths = self.preencoder(feats, feats_lengths)

        # 3. forward encoder
        encoder_out, encoder_out_lens = \
            self.forward_encoder(feats, feat_length)
        encoder_out = self.mask_output(encoder_out, encoder_out_lens)

        # if self.config.do_postencoder:
        #     encoder_out, encoder_out_lens = self.postencoder(
        #         encoder_out, encoder_out_lens
        #     )

        return encoder_out, encoder_out_lens

    def mask_output(self, feats, feat_length):
        if self.config.is_vggrnn:
            feats = mask_fill(feats, make_pad_mask(feat_length, feats, 1), 0.0)
        return feats, feat_length

    def forward_encoder(self, feats, feat_length):
        encoder_out, encoder_out_lens = \
            self.encoder.run(["encoder_out", "encoder_out_lens"], {
                "feats": feats,
                "feats_length": feat_length
            })
        
        if self.config.enc_type == 'RNNEncoder':
            encoder_out = mask_fill(encoder_out, make_pad_mask(
                feat_length, encoder_out, 1), 0.0)

        return encoder_out, encoder_out_lens




from typing import (
    Optional,
    Tuple
)

import math
import torch
import torch.nn as nn

from espnet_onnx.utils.abs_model import AbsExportModel


class OnnxMelGANVocoder(nn.Module, AbsExportModel):
    def __init__(
        self,
        model,
        use_z=False,
        **kwargs
    ):
        super().__init__()
        self.model = model
        self.in_channels = model.melgan[1].in_channels
        self.model_name = 'MelGANVocoder'

    def forward(self, c: torch.Tensor) -> torch.Tensor:
        c = self.model.melgan(c.transpose(1, 0).unsqueeze(0))
        return c.squeeze(0).transpose(1, 0)

    def get_dummy_inputs(self):
        c = torch.randn(100, self.in_channels)
        return (c,)

    def get_input_names(self):
        return ['c']

    def get_output_names(self):
        return ['wav']

    def get_dynamic_axes(self):
        return {
            'c': {0: 'c_length'},
        }

    def get_model_config(self, path):
        return {
            'vocoder_type': 'MelGANVocoder',
            'model_path': str(path / f'{self.model_name}.onnx')
        }



import os
import json
import yaml
import warnings
from pathlib import Path


def get_config(path):
    _, ext = os.path.splitext(path)
    if ext == '.json':
        with open(path, 'r', encoding='utf-8') as f:
            dic = json.load(f)
    elif ext in ('.yaml', '.yml'):
        with open(path, 'r', encoding='utf-8') as f:
            dic = yaml.safe_load(f)
    else:
        raise ValueError('Configuration format is not supported.')
    return Config(dic)

def save_config(config, path):
    _, ext = os.path.splitext(path)
    if ext == '.json':
        with open(path, 'w', encoding='utf-8') as f:
            if isinstance(config, Config):
                f.write(json.dumps(config.__dict__))
            else:
                f.write(json.dumps(config))
    elif ext in ('.yaml', '.yml'):
        with open(path, 'w', encoding='utf-8') as f:
            if isinstance(config, Config):
                yaml.dump(config.__dict__, f)
            else:
                yaml.dump(config, f)
    else:
        raise ValueError(f'File type {ext} is not supported.')

def update_model_path(tag_name, model_path):
    # get configuration of the tag name.
    tag_config_path = Path.home() / ".cache" / "espnet_onnx" / 'tag_config.yaml'
    if os.path.exists(tag_config_path):
        config = get_config(tag_config_path).dic
    else:
        config = {}
    if tag_name in config.keys():
        warnings.warn(f'Onnx model "{tag_name}" is already saved in {config[tag_name]}. ' \
                      + f'Update model path to "{model_path}".')
    config[tag_name] = str(model_path)
    save_config(config, tag_config_path)

def get_tag_config():
    tag_config_path = Path.home() / ".cache" / "espnet_onnx" / 'tag_config.yaml'
    if os.path.exists(tag_config_path):
        config = get_config(tag_config_path)
    else:
        config = {}
    return config

class Config(object):
    def __init__(self, dic=None):
        if dic is not None:
            for j, k in dic.items():
                if isinstance(k, dict):
                    setattr(self, j, Config(k))
                elif isinstance(k, list) and len(k) > 0:
                    if isinstance(k[0], dict):
                        setattr(self, j, [Config(el) for el in k])
                    else:
                        setattr(self, j, k)
                else:
                    if k is not None:
                        setattr(self, j, k)
                    else:
                        setattr(self, j, None)
        self.dic = dic

    def __len__(self):
        return len(self.__dict__.keys())

    def __setitem__(self, key, value):
        setattr(self, key, value)

    def __getitem__(self, key):
        return getattr(self, key)

    def __str__(self):
        return '\n'.join(['%s : %s' % (str(k), str(v)) for k, v in self.__dict__.items()])

    def keys(self):
        return self.__dict__.keys()

    def values(self):
        return self.__dict__.values()



from typing import Tuple
from typeguard import check_argument_types

import numpy as np
import librosa

from espnet_onnx.utils.function import (
    make_pad_mask,
    mask_fill
)
from espnet_onnx.utils.config import Config


class LogMel:
    """Convert STFT to fbank feats
    The arguments is same as librosa.filters.mel
    Args:
        config.fs: number > 0 [scalar] sampling rate of the incoming signal
        config.n_fft: int > 0 [scalar] number of FFT components
        config.n_mels: int > 0 [scalar] number of Mel bands to generate
        config.fmin: float >= 0 [scalar] lowest frequency (in Hz)
        config.fmax: float >= 0 [scalar] highest frequency (in Hz).
            If `None`, use `fmax = fs / 2.0`
        config.htk: use HTK formula instead of Slaney
    """

    def __init__(
        self,
        config: Config
    ):
        fmin = 0 if config.fmin is None else config.fmin
        fmax = fs / 2 if config.fmax is None else config.fmax
        _mel_options = dict(
            sr=config.sr,
            n_fft=config.n_fft,
            n_mels=config.n_mels,
            fmin=fmin,
            fmax=fmax,
            htk=config.htk,
        )
        self.mel_options = _mel_options
        self.log_base = config.log_base
        melmat = librosa.filters.mel(**_mel_options)
        self.melmat = melmat.T

    def extra_repr(self):
        return ", ".join(f"{k}={v}" for k, v in self.mel_options.items())

    def __call__(
        self,
        feat: np.ndarray,
        ilens: np.ndarray = None,
    ) -> Tuple[np.ndarray, np.ndarray]:
        # feat: (B, T, D1) x melmat: (D1, D2) -> mel_feat: (B, T, D2)
        mel_feat = np.matmul(feat, self.melmat)
        mel_feat = np.clip(mel_feat, 1e-10, float('inf'))

        if self.log_base is None:
            logmel_feat = np.log(mel_feat)
        elif self.log_base == 2.0:
            logmel_feat = np.log2(mel_feat)
        elif self.log_base == 10.0:
            logmel_feat = np.log10(mel_feat)
        else:
            logmel_feat = np.log(mel_feat) / np.log(self.log_base)

        # Zero padding
        logmel_feat = mask_fill(
            logmel_feat, make_pad_mask(ilens, logmel_feat, 1), 0.0)
        return logmel_feat, ilens



from typing import Iterable
from typing import List
from typing import Union
from pathlib import Path
from typeguard import check_argument_types

import sentencepiece as spm


class SentencepiecesTokenizer:
    def __init__(self, model: Union[Path, str]):
        assert check_argument_types()
        self.model = str(model)
        # NOTE(kamo):
        # Don't build SentencePieceProcessor in __init__()
        # because it's not picklable and it may cause following error,
        # "TypeError: can't pickle SwigPyObject objects",
        # when giving it as argument of "multiprocessing.Process()".
        self.sp = None

    def __repr__(self):
        return f'{self.__class__.__name__}(model="{self.model}")'

    def _build_sentence_piece_processor(self):
        # Build SentencePieceProcessor lazily.
        if self.sp is None:
            self.sp = spm.SentencePieceProcessor()
            self.sp.load(self.model)

    def text2tokens(self, line: str) -> List[str]:
        self._build_sentence_piece_processor()
        return self.sp.EncodeAsPieces(line)

    def tokens2text(self, tokens: Iterable[str]) -> str:
        self._build_sentence_piece_processor()
        return self.sp.DecodePieces(list(tokens))




from .tts_models.vits import VITS
from .tts_models.fast_speech2 import FastSpeech2
from .tts_models.tacotron2 import Tacotron2


def get_tts_model(config, providers, use_quantized):
    if config.model_type == 'VITS':
        return VITS(config, providers, use_quantized)
    elif config.model_type == 'FastSpeech2':
        return FastSpeech2(config, providers, use_quantized)
    elif config.model_type == 'Tacotron2':
        return Tacotron2(config, providers, use_quantized)



# -*- coding: utf-8 -*-
"""
Created on Sat Aug 21 17:27:16 2021.
@author: Keqi Deng (UCAS)

Modified by Masao Someki
"""
import os
from typing import (
    Optional,  # noqa: H301
    Tuple,  # noqa: H301
)

import torch
import torch.nn as nn
from espnet.nets.pytorch_backend.transformer.subsampling_without_posenc import (
    Conv2dSubsamplingWOPosEnc,  # noqa: H301
)
from espnet2.asr.frontend.default import DefaultFrontend
from espnet2.layers.global_mvn import GlobalMVN
from espnet2.layers.utterance_mvn import UtteranceMVN

from espnet_onnx.utils.abs_model import AbsExportModel


class ContextualBlockXformerEncoder(nn.Module, AbsExportModel):
    """Contextual Block Conformer encoder module.
    """
    def __init__(
        self,
        model,
        feats_dim=80,
        **kwargs
    ):
        super().__init__()
        self.model = model
        self.model_name = 'xformer_encoder'
        self._output_size = model._output_size
        self.pos_enc = model.pos_enc
        
        self.embed = model.embed
        self.subsample = model.subsample
        
        self.normalize_before = model.normalize_before
        self.encoders = model.encoders
        
        if self.normalize_before:
            self.after_norm = model.after_norm

        # for block processing
        self.block_size = model.block_size
        self.hop_size = model.hop_size
        self.look_ahead = model.look_ahead
        self.init_average = model.init_average
        self.ctx_pos_enc = model.ctx_pos_enc
        self.overlap_size = self.block_size - self.hop_size
        self.offset = self.block_size - self.look_ahead - self.hop_size
        self.xscale = model.pos_enc.xscale
        
        # for export configuration
        self.feats_dim = feats_dim

    def output_size(self) -> int:
        return self._output_size

    def forward(
        self,
        xs_pad: torch.Tensor,
        mask: torch.Tensor,
        buffer_before_downsampling: torch.Tensor,
        buffer_after_downsampling: torch.Tensor,
        prev_addin: torch.Tensor,
        pos_enc_xs: torch.Tensor,
        pos_enc_addin: torch.Tensor,
        past_encoder_ctx: torch.Tensor,
        indicies: torch.Tensor,
    ) -> Tuple[torch.Tensor, torch.Tensor, Optional[torch.Tensor]]:
        """
        The sum of the length xs_pad:L_1 and of buffer_before_downsampling: L_2
        should be the multiple of subsample.
        L_1 + L_2 = alpha * self.subsample
        
        Args:
            xs_pad: (1, hop_size*subsample+1, D)
            buffer_after_downsampling: (1, overlap_size=block_size-hop_size, D)
            indicies: torch.Tensor. [offset, offset + hop_length, overlap_size]
            mask: zeros(1, 1, self.block_size + 2, self.block_size + 2)
            pos_enc_xs: (B, L, D) L = block_size
        """
        xs_pad = torch.cat([buffer_before_downsampling, xs_pad], dim=1)
        buffer_before_downsampling = xs_pad[:, -self.subsample:] # (B, L, overlap)
        xs_pad = self.compute_embed(xs_pad)
        xs_pad = torch.cat([buffer_after_downsampling, xs_pad], dim=1)
        
        buffer_after_downsampling = xs_pad[:, -indicies[2]:] # (B, L, overlap)
        
        if self.init_average:
            addin = xs_pad.mean(1, keepdim=True)
        else:
            addin = xs_pad.max(1, keepdim=True)
            
        if self.ctx_pos_enc:
            addin = addin * self.xscale + pos_enc_addin 
        
        xs_pad = xs_pad * self.xscale + pos_enc_xs 
        xs_chunk = torch.cat([prev_addin, xs_pad, addin], dim=1)

        ys_chunk, _, _, _, past_encoder_ctx, _, _ = self.encoders(
            xs_chunk.unsqueeze(1), mask, True, past_encoder_ctx
        )

        # remove addin
        ys_chunk = ys_chunk.squeeze(1)[:, 1:-1]
        ys_pad = ys_chunk[:, indicies[0] : indicies[1]]

        if self.normalize_before:
            ys_pad = self.after_norm(ys_pad)

        return ys_pad, buffer_before_downsampling, buffer_after_downsampling, addin, past_encoder_ctx
    
    def compute_embed(self, xs_pad):
        if isinstance(self.embed, Conv2dSubsamplingWOPosEnc):
            xs_pad, _ = self.embed(xs_pad, None)
        elif self.embed is not None:
            xs_pad = self.embed(xs_pad)
        return xs_pad

    def get_output_size(self):
        return self._output_size

    def get_dummy_inputs(self):
        n_feats = self.feats_dim
        xs_pad = torch.randn(1, self.hop_size*self.subsample, n_feats)
        mask = torch.ones(1, 1, self.block_size + 2, self.block_size + 2)
        o = self.compute_embed(xs_pad)
        buffer_before_downsampling = torch.randn(1, self.subsample, n_feats)
        buffer_after_downsampling = torch.randn(1, self.overlap_size, o.shape[-1])
        prev_addin = torch.randn(1, 1, o.shape[-1])
        pos_enc_xs = torch.randn(1, self.block_size, o.shape[-1])
        pos_enc_addin = torch.randn(1, 1, o.shape[-1])
        past_encoder_ctx = torch.randn(1, len(self.encoders), self.encoders[0].size)
        indicies = torch.LongTensor([8, 24, 24])
        return (xs_pad, mask, buffer_before_downsampling, buffer_after_downsampling, prev_addin,
                pos_enc_xs, pos_enc_addin, past_encoder_ctx, indicies)

    def get_input_names(self):
        return ['xs_pad', 'mask', 'buffer_before_downsampling', 'buffer_after_downsampling',
                'prev_addin', 'pos_enc_xs', 'pos_enc_addin', 'past_encoder_ctx', 'indicies']

    def get_output_names(self):
        return ['ys_pad', 'next_buffer_before_downsampling', 'next_buffer_after_downsampling',
                'next_addin', 'next_encoder_ctx']

    def get_dynamic_axes(self):
        return {
            'xs_pad': { 1: 'xs_pad_length' },
            'mask': { 2: 'block_height', 3: 'block_width' },
            'buffer_before_downsampling': { 1: 'bbd_length' },
            'buffer_after_downsampling': { 1: 'bad_length' },
            'pos_enc_xs': { 1: 'pex_length' },
            'ys_pad': { 1: 'ys_pad_length' },
            'next_buffer_before_downsampling': { 1: 'nbbd_length' },
            'next_buffer_after_downsampling': { 1: 'nbad_length' },
        }

    def get_model_config(self, asr_model=None, path=None):
        ret = {}
        ret.update(
            enc_type='ContextualXformerEncoder',
            model_path=os.path.join(path, f'{self.model_name}.onnx'),
            frontend=self.get_frontend_config(asr_model.frontend),
            do_normalize=asr_model.normalize is not None,
            do_preencoder=asr_model.preencoder is not None,
            do_postencoder=asr_model.postencoder is not None
        )
        if ret['do_normalize']:
            ret.update(normalize=self.get_norm_config(
                asr_model.normalize, path))
        # streaming config
        ret.update(
            pe_path=str(path.parent / 'pe.npy'),
            n_layers=len(asr_model.encoder.encoders),
            subsample=self.subsample,
        )
        # Currently preencoder, postencoder is not supported.
        # if ret['do_preencoder']:
        #     ret.update(preencoder=get_preenc_config(self.model.preencoder))
        # if ret['do_postencoder']:
        #     ret.update(postencoder=get_postenc_config(self.model.postencoder))
        return ret

    def get_frontend_config(self, frontend):
        # currently only default config is supported.
        assert isinstance(
            frontend, DefaultFrontend), 'Currently only DefaultFrontend is supported.'

        stft_config = dict(
            n_fft=frontend.stft.n_fft,
            win_length=frontend.stft.win_length,
            hop_length=frontend.stft.hop_length,
            window=frontend.stft.window,
            center=False, # Set false to compute stft continuously
            onesided=frontend.stft.onesided,
            normalized=frontend.stft.normalized,
        )
        logmel_config = frontend.logmel.mel_options
        logmel_config.update(log_base=frontend.logmel.log_base)
        return {
            "stft": stft_config,
            "logmel": logmel_config
        }

    def get_norm_config(self, normalize, path):
        if isinstance(normalize, GlobalMVN):
            return {
                "type": "gmvn",
                "norm_means": normalize.norm_means,
                "norm_vars": normalize.norm_vars,
                "eps": normalize.eps,
                "stats_file": str(path.parent / 'feats_stats.npz')
            }
        elif isinstance(normalize, UtteranceMVN):
            return {
                "type": "utterance_mvn",
                "norm_means": normalize.norm_means,
                "norm_vars": normalize.norm_vars,
                "eps": normalize.eps,
            }


"""Parallel beam search module."""

import logging
from typing import (
    Any,
    Dict,
    List,
    Tuple
)
import numpy as np

from espnet_onnx.utils.function import (
    pad_sequence,
    topk
)
from .beam_search import BeamSearch
from .hyps import Hypothesis
from .hyps import BatchHypothesis


class BatchBeamSearch(BeamSearch):
    """Batch beam search implementation."""

    def batchfy(self, hyps: List[Hypothesis]) -> BatchHypothesis:
        """Convert list to batch."""
        if len(hyps) == 0:
            return BatchHypothesis()
        return BatchHypothesis(
            yseq=pad_sequence(
                [h.yseq for h in hyps], batch_first=True, padding_value=self.eos
            ),
            length=np.array([len(h.yseq) for h in hyps], dtype=np.int64),
            score=np.array([h.score for h in hyps]),
            scores={k: np.array([h.scores[k] for h in hyps])
                    for k in self.scorers},
            states={k: [h.states[k] for h in hyps] for k in self.scorers},
        )

    def _batch_select(self, hyps: BatchHypothesis, ids: List[int]) -> BatchHypothesis:
        return BatchHypothesis(
            yseq=hyps.yseq[ids],
            score=hyps.score[ids],
            length=hyps.length[ids],
            scores={k: v[ids] for k, v in hyps.scores.items()},
            states={
                k: [self.scorers[k].select_state(v, i) for i in ids]
                for k, v in hyps.states.items()
            },
        )

    def _select(self, hyps: BatchHypothesis, i: int) -> Hypothesis:
        return Hypothesis(
            yseq=hyps.yseq[i, : hyps.length[i]],
            score=hyps.score[i],
            scores={k: v[i] for k, v in hyps.scores.items()},
            states={
                k: self.scorers[k].select_state(v, i) for k, v in hyps.states.items()
            },
        )

    def unbatchfy(self, batch_hyps: BatchHypothesis) -> List[Hypothesis]:
        """Revert batch to list."""
        return [
            Hypothesis(
                yseq=batch_hyps.yseq[i][: batch_hyps.length[i]],
                score=batch_hyps.score[i],
                scores={k: batch_hyps.scores[k][i] for k in self.scorers},
                states={
                    k: v.select_state(batch_hyps.states[k], i)
                    for k, v in self.scorers.items()
                },
            )
            for i in range(len(batch_hyps.length))
        ]

    def batch_beam(
        self, weighted_scores: np.ndarray, ids: np.ndarray
    ) -> Tuple[np.ndarray, np.ndarray, np.ndarray, np.ndarray]:
        """Batch-compute topk full token ids and partial token ids.
        Args:
            weighted_scores (np.ndarray): The weighted sum scores for each tokens.
                Its shape is `(n_beam, self.vocab_size)`.
            ids (np.ndarray): The partial token ids to compute topk.
                Its shape is `(n_beam, self.pre_beam_size)`.
        Returns:
            Tuple[np.ndarray, np.ndarray, np.ndarray, np.ndarray]:
                The topk full (prev_hyp, new_token) ids
                and partial (prev_hyp, new_token) ids.
                Their shapes are all `(self.beam_size,)`
        """
        top_ids = topk(weighted_scores.reshape(-1), self.beam_size)
        prev_hyp_ids = top_ids // self.n_vocab
        new_token_ids = top_ids % self.n_vocab
        return prev_hyp_ids, new_token_ids, prev_hyp_ids, new_token_ids

    def init_hyp(self, x: np.ndarray) -> BatchHypothesis:
        """Get an initial hypothesis data.
        Args:
            x (np.ndarray): The encoder output feature
        Returns:
            Hypothesis: The initial hypothesis.
        """
        init_states = dict()
        init_scores = dict()
        for k, d in self.scorers.items():
            init_states[k] = d.batch_init_state(x)
            init_scores[k] = 0.0
        return self.batchfy(
            [
                Hypothesis(
                    score=0.0,
                    scores=init_scores,
                    states=init_states,
                    yseq=np.array([0, self.sos], dtype=np.int64),
                )
            ]
        )

    def score_full(
        self, hyp: BatchHypothesis, x: np.ndarray
    ) -> Tuple[Dict[str, np.ndarray], Dict[str, Any]]:
        """Score new hypothesis by `self.full_scorers`.
        Args:
            hyp (Hypothesis): Hypothesis with prefix tokens to score
            x (np.ndarray): Corresponding input feature
        Returns:
            Tuple[Dict[str, np.ndarray], Dict[str, Any]]: Tuple of
                score dict of `hyp` that has string keys of `self.full_scorers`
                and tensor score values of shape: `(self.n_vocab,)`,
                and state dict that has string keys
                and state values of `self.full_scorers`
        """
        scores = dict()
        states = dict()
        for k, d in self.full_scorers.items():
            scores[k], states[k] = d.batch_score(hyp.yseq, hyp.states[k], x)
        return scores, states

    def score_partial(
        self, hyp: BatchHypothesis, ids: np.ndarray, x: np.ndarray
    ) -> Tuple[Dict[str, np.ndarray], Dict[str, Any]]:
        """Score new hypothesis by `self.full_scorers`.
        Args:
            hyp (Hypothesis): Hypothesis with prefix tokens to score
            ids (np.ndarray): 2D tensor of new partial tokens to score
            x (np.ndarray): Corresponding input feature
        Returns:
            Tuple[Dict[str, np.ndarray], Dict[str, Any]]: Tuple of
                score dict of `hyp` that has string keys of `self.full_scorers`
                and tensor score values of shape: `(self.n_vocab,)`,
                and state dict that has string keys
                and state values of `self.full_scorers`
        """
        scores = dict()
        states = dict()
        for k, d in self.part_scorers.items():
            scores[k], states[k] = d.batch_score_partial(
                hyp.yseq, ids, hyp.states[k], x
            )
        return scores, states

    def merge_states(self, states: Any, part_states: Any, part_idx: int) -> Any:
        """Merge states for new hypothesis.
        Args:
            states: states of `self.full_scorers`
            part_states: states of `self.part_scorers`
            part_idx (int): The new token id for `part_scores`
        Returns:
            Dict[str, np.ndarray]: The new score dict.
                Its keys are names of `self.full_scorers` and `self.part_scorers`.
                Its values are states of the scorers.
        """
        new_states = dict()
        for k, v in states.items():
            new_states[k] = v
        for k, v in part_states.items():
            new_states[k] = v
        return new_states

    def search(self, running_hyps: BatchHypothesis, x: np.ndarray) -> BatchHypothesis:
        """Search new tokens for running hypotheses and encoded speech x.
        Args:
            running_hyps (BatchHypothesis): Running hypotheses on beam
            x (np.ndarray): Encoded speech feature (T, D)
        Returns:
            BatchHypothesis: Best sorted hypotheses
        """
        n_batch = len(running_hyps)
        part_ids = None  # no pre-beam
        # batch scoring
        weighted_scores = np.zeros(
            (n_batch, self.n_vocab), dtype=x.dtype
        )
        scores, states = self.score_full(running_hyps, np.vstack(
            [x for _ in range(n_batch)]).reshape(n_batch, *x.shape))
        for k in self.full_scorers:
            weighted_scores += self.weights[k] * scores[k]
        # partial scoring
        if self.do_pre_beam:
            pre_beam_scores = (
                weighted_scores
                if self.pre_beam_score_key == "full"
                else scores[self.pre_beam_score_key]
            )
            part_ids = topk(pre_beam_scores, self.pre_beam_size)
        part_scores, part_states = self.score_partial(
            running_hyps, part_ids, x)
        for k in self.part_scorers:
            weighted_scores += self.weights[k] * part_scores[k]
        # add previous hyp scores
        weighted_scores += np.expand_dims(running_hyps.score, 1)

        best_hyps = []
        prev_hyps = self.unbatchfy(running_hyps)
        for (
            full_prev_hyp_id,
            full_new_token_id,
            part_prev_hyp_id,
            part_new_token_id,
        ) in zip(*self.batch_beam(weighted_scores, part_ids)):
            prev_hyp = prev_hyps[full_prev_hyp_id]
            best_hyps.append(
                Hypothesis(
                    score=weighted_scores[full_prev_hyp_id, full_new_token_id],
                    yseq=self.append_token(prev_hyp.yseq, full_new_token_id),
                    scores=self.merge_scores(
                        prev_hyp.scores,
                        {k: v[full_prev_hyp_id] for k, v in scores.items()},
                        full_new_token_id,
                        {k: v[part_prev_hyp_id]
                            for k, v in part_scores.items()},
                        part_new_token_id,
                    ),
                    states=self.merge_states(
                        {
                            k: self.full_scorers[k].select_state(
                                v, full_prev_hyp_id)
                            for k, v in states.items()
                        },
                        {
                            k: self.part_scorers[k].select_state(
                                v, part_prev_hyp_id, part_new_token_id
                            )
                            for k, v in part_states.items()
                        },
                        part_new_token_id,
                    ),
                )
            )
        return self.batchfy(best_hyps)

    def post_process(
        self,
        i: int,
        maxlen: int,
        running_hyps: BatchHypothesis,
        ended_hyps: List[Hypothesis],
    ) -> BatchHypothesis:
        """Perform post-processing of beam search iterations.
        Args:
            i (int): The length of hypothesis tokens.
            maxlen (int): The maximum length of tokens in beam search.
            running_hyps (BatchHypothesis): The running hypotheses in beam search.
            ended_hyps (List[Hypothesis]): The ended hypotheses in beam search.
        Returns:
            BatchHypothesis: The new running hypotheses.
        """
        n_batch = running_hyps.yseq.shape[0]
        logging.debug(f"the number of running hypothes: {n_batch}")
        if self.token_list is not None:
            logging.debug(
                "best hypo: "
                + "".join(
                    [
                        self.token_list[int(x)]
                        for x in running_hyps.yseq[0, 1: running_hyps.length[0]]
                    ]
                )
            )
        # add eos in the final loop to avoid that there are no ended hyps
        if i == maxlen - 1:
            logging.debug("adding <eos> in the last position in the loop")
            yseq_eos = np.hstack(
                (
                    running_hyps.yseq,
                    np.full(
                        (n_batch, 1),
                        self.eos,
                        dtype=np.int64,
                    ),
                )
            )

            running_hyps.yseq.resize(yseq_eos.shape, refcheck=False)
            running_hyps.yseq[:] = yseq_eos
            running_hyps.length[:] = yseq_eos.shape[1]

        # add ended hypotheses to a final list, and removed them from current hypotheses
        # (this will be a probmlem, number of hyps < beam)
        is_eos = (
            running_hyps.yseq[np.arange(
                n_batch), (running_hyps.length - 1).astype(np.int64)]
            == self.eos
        )
        for b in np.transpose(np.nonzero(is_eos)).reshape(-1):
            hyp = self._select(running_hyps, b)
            ended_hyps.append(hyp)
        remained_ids = np.transpose(np.nonzero(is_eos == 0)).reshape(-1)
        return self._batch_select(running_hyps, remained_ids)



# -*- coding: utf-8 -*-

# Copyright 2020 Nagoya University (Tomoki Hayashi)
#  Apache 2.0  (http://www.apache.org/licenses/LICENSE-2.0)
# This script is copied from espnet2.tts.utils.duration_calculator and modified by Masao Someki

"""Duration calculator for ESPnet2."""

from typing import Tuple
import numpy as np


class DurationCalculator:
    """Duration calculator module."""

    def __call__(self, att_ws: np.ndarray) -> Tuple[np.ndarray, np.ndarray]:
        """Convert attention weight to durations.
        Args:
            att_ws (Tesnor): Attention weight tensor (T_feats, T_text) or
                (#layers, #heads, T_feats, T_text).
        Returns:
            LongTensor: Duration of each input (T_text,).
            Tensor: Focus rate value.
        """
        duration = self._calculate_duration(att_ws)
        focus_rate = self._calculate_focus_rete(att_ws)

        return duration, focus_rate

    @staticmethod
    def _calculate_focus_rete(att_ws):
        if len(att_ws.shape) == 2:
            # tacotron 2 case -> (T_feats, T_text)
            return att_ws.max(axis=-1)[0].mean()
        elif len(att_ws.shape) == 4:
            # transformer case -> (#layers, #heads, T_feats, T_text)
            return att_ws.max(axis=-1)[0].mean(axis=-1).max()
        else:
            raise ValueError("att_ws should be 2 or 4 dimensional tensor.")

    @staticmethod
    def _calculate_duration(att_ws):
        if len(att_ws.shape) == 2:
            # tacotron 2 case -> (T_feats, T_text)
            pass
        elif len(att_ws.shape) == 4:
            # transformer case -> (#layers, #heads, T_feats, T_text)
            # get the most diagonal head according to focus rate
            att_ws = np.concatenate(
                [att_w for att_w in att_ws], axis=0
            )  # (#heads * #layers, T_feats, T_text)
            diagonal_scores = att_ws.max(
                axis=-1)[0].mean(axis=-1)  # (#heads * #layers,)
            diagonal_head_idx = diagonal_scores.argmax()
            att_ws = att_ws[diagonal_head_idx]  # (T_feats, T_text)
        else:
            raise ValueError("att_ws should be 2 or 4 dimensional tensor.")
        # calculate duration from 2d attention weight
        durations = np.stack(
            [(att_ws.argmax(-1) == i).sum() for i in range(att_ws.shape[1])]
        )
        return durations.reshape(-1)



from typing import (
    Dict,
    List,
    Tuple
)

import onnxruntime
import numpy as np

from espnet_onnx.asr.frontend.frontend import Frontend
from espnet_onnx.asr.frontend.global_mvn import GlobalMVN
from espnet_onnx.asr.frontend.utterance_mvn import UtteranceMVN
from espnet_onnx.utils.config import Config


class StreamingEncoder:
    def __init__(
        self,
        encoder_config: Config,
        providers: List[str],
        use_quantized: bool = False,
    ):
        self.config = encoder_config
        if use_quantized:
            self.encoder = onnxruntime.InferenceSession(
                self.config.quantized_model_path,
                providers=providers
                )
        else:
            self.encoder = onnxruntime.InferenceSession(
                self.config.model_path,
                providers=providers
                )

        self.frontend = Frontend(self.config.frontend)
        if self.config.do_normalize:
            if self.config.normalize.type == 'gmvn':
                self.normalize = GlobalMVN(self.config.normalize)
            elif self.config.normalize.type == 'utterance_mvn':
                self.normalize = UtteranceMVN(self.config.normalize)
        
        self.pos_enc = np.load(self.config.pe_path)
        self.n_processed_blocks = 0
        self.n_layers = self.config.n_layers
        self.overlap_size = self.config.block_size - self.config.hop_size
        self.outputs = ['ys_pad','next_buffer_before_downsampling',
            'next_buffer_after_downsampling', 'next_addin', 'next_encoder_ctx']

        # if self.config.do_preencoder:
        #     self.preencoder = Preencoder(self.config.preencoder)
        # if self.config.do_postencoder:
        #     self.postencoder = Postencoder(self.config.postencoder)

    def __call__(
        self, speech: np.ndarray, speech_length: np.ndarray, states
    ) -> Tuple[np.ndarray, np.ndarray]:
        """Frontend + Encoder. Note that this method is used by asr_inference.py
        Args:
            speech: (Batch, Length, ...)
            speech_lengths: (Batch, )
        """
        # 1. Extract feature
        feats, feat_length = self.frontend(speech, speech_length)
        
        # 2. normalize with global MVN
        if self.config.do_normalize:
            feats, feat_length = self.normalize(feats, feat_length)

        # if self.config.do_preencoder:
        #     feats, feats_lengths = self.preencoder(feats, feats_lengths)

        # 3. forward encoder
        encoder_out, next_states = self.forward_encoder(feats, states)

        # if self.config.do_postencoder:
        #     encoder_out, encoder_out_lens = self.postencoder(
        #         encoder_out, encoder_out_lens
        #     )

        return encoder_out, next_states

    def forward_encoder(self, feats: np.ndarray, states: Dict[str, np.ndarray]):
        input_dict = self.get_input_dict(feats, states)
        ys_pad, nbbd, nbad, naddin, nec = self.encoder.run(self.outputs, input_dict)
        ret = {
            'buffer_before_downsampling' : nbbd,
            'buffer_after_downsampling' : nbad,
            'prev_addin' : naddin,
            'past_encoder_ctx' : nec,
        }
        return ys_pad, ret

    def forward_final(self, states: Dict[str, np.ndarray]):
        dummy_xs_pad = np.zeros(
            (1, self.config.hop_size * self.config.subsample + 1, self.config.frontend.logmel.n_mels),
            dtype=np.float32
        )
        input_dict = self.get_input_dict(dummy_xs_pad, states)
        ys_pad, *_ = self.encoder.run(self.outputs, input_dict)
        return ys_pad, None

    def get_input_dict(self, x: np.ndarray, state: Dict[str, np.ndarray]):
        # x.length : hop_size * subsample + 1
        mask = np.zeros(
            (1, 1, self.config.block_size + 2, self.config.block_size + 2),
            dtype=np.float32
        )
        mask[..., 1:, :-1] = 1
        if self.n_processed_blocks == 0:
            start = 0
            indicies = np.array([
                0,
                self.config.block_size - self.config.look_ahead,
                self.overlap_size
            ], dtype=np.int64)
        else:
            start = self.config.hop_size * self.n_processed_blocks
            offset = self.config.block_size - self.config.look_ahead - self.config.hop_size + 1
            indicies = np.array([
                offset,
                offset + self.config.hop_size,
                self.overlap_size
            ], dtype=np.int64)
        return {
            'xs_pad': x[:, :-1],
            'mask': mask,
            'buffer_before_downsampling': state['buffer_before_downsampling'],
            'buffer_after_downsampling': state['buffer_after_downsampling'],
            'prev_addin': state['prev_addin'],
            'pos_enc_xs': self.pos_enc[:, start : start + self.config.block_size],
            'pos_enc_addin': self.pos_enc[:, self.n_processed_blocks : self.n_processed_blocks + 1],
            'past_encoder_ctx': state['past_encoder_ctx'],
            'indicies': indicies
        }
    
    def reset(self):
        self.n_processed_blocks = 0
    
    def increment(self):
        self.n_processed_blocks += 1
    
    def init_state(self):
        return {
            'buffer_before_downsampling' : np.zeros((1, self.config.subsample, self.config.frontend.logmel.n_mels), dtype=np.float32),
            'buffer_after_downsampling' : np.zeros((1, self.config.block_size - self.config.hop_size, self.pos_enc.shape[-1]), dtype=np.float32),
            'prev_addin' : np.zeros((1, 1, self.pos_enc.shape[-1]), dtype=np.float32),
            'past_encoder_ctx' : np.zeros((1, self.n_layers, self.pos_enc.shape[-1]), dtype=np.float32),
        }



import os

import torch
import torch.nn as nn

from espnet2.asr.decoder.transformer_decoder import TransformerDecoder

from espnet_onnx.utils.function import subsequent_mask
from ..language_models.embed import Embedding
from espnet_onnx.utils.abs_model import AbsExportModel


class XformerDecoder(nn.Module, AbsExportModel):
    def __init__(self, model, max_seq_len=512, **kwargs):
        super().__init__()
        self.embed = Embedding(model.embed, max_seq_len)
        self.model = model
        self.model_name = 'xformer_decoder'

    def forward(self, tgt, tgt_mask, memory, cache):
        x = self.embed(tgt)
        new_cache = []
        for c, decoder in zip(cache, self.model.decoders):
            x, tgt_mask, memory, memory_mask = decoder(
                x, tgt_mask, memory, None, cache=c
            )
            new_cache.append(x)  # (1, L, 512) * n_layer
        y = self.model.after_norm(x[:, -1])
        y = torch.log_softmax(self.model.output_layer(y), dim=-1)
        return y, new_cache

    def get_dummy_inputs(self, enc_size):
        tgt = torch.LongTensor([0, 1]).unsqueeze(0)
        tgt_mask = torch.from_numpy(subsequent_mask(2)[None, :])
        enc_out = torch.randn(1, 100, enc_size)
        cache = [
            torch.zeros((1, 1, self.model.decoders[0].size))
            for _ in range(len(self.model.decoders))
        ]
        return (tgt, tgt_mask, enc_out, cache)

    def get_input_names(self):
        return ['tgt', 'tgt_mask', 'memory'] \
            + ['cache_%d' % i for i in range(len(self.model.decoders))]

    def get_output_names(self):
        return ['y'] \
            + ['out_cache_%d' % i for i in range(len(self.model.decoders))]

    def get_dynamic_axes(self):
        ret = {
            'tgt': {
                0: 'tgt_batch',
                1: 'tgt_length'
            },
            'tgt_mask': {
                1: 'tgt_mask_length',
                2: 'tgt_mask_height'
            },
            'memory': {
                0: 'memory_batch',
                1: 'memory_length'
            }
        }
        ret.update({
            'cache_%d' % d: {
                0: 'cache_%d_batch' % d,
                1: 'cache_%d_length' % d
            }
            for d in range(len(self.model.decoders))
        })
        ret.update({
            'out_cache_%d' % d: {
                0: 'out_cache_%d_batch' % d,
                1: 'out_cache_%d_length' % d
            }
            for d in range(len(self.model.decoders))
        })
        return ret

    def get_model_config(self, path):
        return {
            "dec_type": "XformerDecoder",
            "model_path": os.path.join(path, f'{self.model_name}.onnx'),
            "n_layers": len(self.model.decoders),
            "odim": self.model.decoders[0].size
        }



from typing import List
from espnet_onnx.utils.config import Config
from espnet_onnx.asr.model.decoders.rnn import RNNDecoder
from espnet_onnx.asr.model.decoders.xformer import XformerDecoder
from espnet_onnx.asr.model.decoders.transducer import TransducerDecoder


def get_decoder(config: Config, providers: List[str], use_quantized: bool = False):
    if config.dec_type == 'RNNDecoder':
        return RNNDecoder(config, providers, use_quantized)
    elif config.dec_type == 'TransducerDecoder':
        return TransducerDecoder(config, providers, use_quantized)
    else:
        return XformerDecoder(config, providers, use_quantized)



from typing import List
from espnet_onnx.utils.config import Config
from espnet_onnx.asr.model.lms.seqrnn_lm import SequentialRNNLM
from espnet_onnx.asr.model.lms.transformer_lm import TransformerLM


def get_lm(config: Config, providers: List[str], use_quantized: bool = False):
    if config.lm.use_lm:
        if config.lm.lm_type == 'SequentialRNNLM':
            return SequentialRNNLM(config.lm, providers, use_quantized)
        elif config.lm.lm_type == 'TransformerLM':
            return TransformerLM(config.lm, providers, use_quantized)
    return None



import os

import torch
import torch.nn as nn

from espnet_onnx.utils.abs_model import AbsExportModel


class SequentialRNNLM(nn.Module, AbsExportModel):
    def __init__(self, model, **kwargs):
        super().__init__()
        self.encoder = model.encoder
        self.rnn = model.rnn
        self.rnn_type = model.rnn_type
        self.decoder = model.decoder
        self.nlayers = model.nlayers
        self.nhid = model.nhid
        self.model_name = 'seq_rnnlm'

    def forward(self, y, hidden1, hidden2=None):
        # batch_score function.
        emb = self.encoder(y)
        if self.rnn_type == 'LSTM':
            output, (hidden1, hidden2) = self.rnn(emb, (hidden1, hidden2))
        else:
            output, hidden1 = self.rnn(emb, hidden1)

        decoded = self.decoder(
            output.contiguous().view(output.size(0) * output.size(1), output.size(2))
        )
        if self.rnn_type == 'LSTM':
            return (
                decoded.view(output.size(0), output.size(1), decoded.size(1)),
                hidden1,
                hidden2
            )
        else:
            return (
                decoded.view(output.size(0), output.size(1), decoded.size(1)),
                hidden1
            )

    def get_dummy_inputs(self):
        tgt = torch.LongTensor([0, 1]).unsqueeze(0)
        hidden = torch.randn(self.nlayers, 1, self.nhid)
        if self.rnn_type == 'LSTM':
            return (tgt, hidden, hidden)
        else:
            return (tgt, hidden)

    def get_input_names(self):
        if self.rnn_type == 'LSTM':
            return ['x', 'in_hidden1', 'in_hidden2']
        else:
            return ['x', 'in_hidden1']

    def get_output_names(self):
        if self.rnn_type == 'LSTM':
            return ['y', 'out_hidden1', 'out_hidden2']
        else:
            return ['y', 'out_hidden1']

    def get_dynamic_axes(self):
        ret = {
            'x': {
                0: 'x_batch',
                1: 'x_length'
            },
            'y': {
                0: 'y_batch'
            },
            'in_hidden1': {
                1: 'hidden1_batch'
            },
            'out_hidden1': {
                1: 'out_hidden1_batch'
            }
        }
        if self.rnn_type == 'LSTM':
            ret.update({
                'in_hidden2': {
                    1: 'hidden2_batch'
                },
                'out_hidden2': {
                    1: 'out_hidden2_batch'
                }
            })
        return ret

    def get_model_config(self, path):
        return {
            "use_lm": True,
            "model_path": os.path.join(path, f'{self.model_name}.onnx'),
            "lm_type": "SequentialRNNLM",
            "rnn_type": self.rnn_type,
            "nhid": self.nhid,
            "nlayers": self.nlayers
        }



from typing import (
    Union,
    List,
    Tuple,
    Optional
)
from pathlib import Path
from typeguard import check_argument_types

import os
import logging
import numpy as np
import glob
import warnings
import onnxruntime

from espnet_onnx.asr.abs_asr_model import AbsASRModel
from espnet_onnx.asr.beam_search.batch_beam_search import BatchBeamSearch
from espnet_onnx.asr.beam_search.batch_beam_search_online_sim import BatchBeamSearchOnlineSim
from espnet_onnx.asr.beam_search.hyps import Hypothesis

from espnet_onnx.utils.config import get_config
from espnet_onnx.utils.config import get_tag_config


class StreamingSpeech2Text(AbsASRModel):
    """Speech2Text class to support streaming asr
    """

    def __init__(self,
                 tag_name: str = None,
                 model_dir: Union[Path, str] = None,
                 providers: List[str] = ['CPUExecutionProvider'],
                 use_quantized: bool = False,
                 block_size: int = 40,
                 hop_size: int = 16,
                 look_ahead: int = 16
                 ):
        assert check_argument_types()
        self._check_argument(tag_name, model_dir)
        self._load_config()

        # check onnxruntime version and providers
        self._check_ort_version(providers)

        # check if model is exported for streaming.
        if self.config.encoder.enc_type != 'ContextualXformerEncoder':
            raise RuntimeError(
                'Onnx model is not build for streaming. Use Speech2Text instead.')

        if use_quantized and 'quantized_model_path' not in self.config.encoder.keys():
            # check if quantized model config is defined.
            raise RuntimeError(
                'Configuration for quantized model is not defined.')

        # build models
        self.config.encoder.block_size = block_size
        self.config.encoder.hop_size = hop_size
        self.config.encoder.look_ahead = look_ahead

        self._build_model(providers, use_quantized)

        # Fix beam_search components
        self.batch_beam_search = BatchBeamSearch(
            self.config.beam_search,
            self.config.token,
            scorers=self.scorers,
            weights=self.weights,
        )
        self.beam_search.__class__ = BatchBeamSearchOnlineSim
        self.beam_search.set_streaming_config(block_size, hop_size, look_ahead)
        logging.info("BatchBeamSearchOnlineSim implementation is selected.")

        # streaming related parameters
        self.enc_feats = []
        self.streaming_states = {
            'buffer_before_downsampling': None,
            'buffer_after_downsampling': None,
            'prev_addin': None,
            'past_encoder_ctx': None,
        }
        self.hop_size = self.config.encoder.frontend.stft.hop_length * self.config.encoder.subsample * hop_size  \
            + (self.config.encoder.frontend.stft.n_fft // self.config.encoder.frontend.stft.hop_length) * \
            self.config.encoder.frontend.stft.hop_length

    def __call__(self, speech: np.ndarray) -> List[
        Tuple[
            Optional[str],
            List[str],
            List[int],
            Union[Hypothesis],
        ]
    ]:
        """Inference

        Args:
            data: Input speech data

        Returns:
            text, token, token_int, hyp

        """
        assert check_argument_types()

        # check dtype
        if speech.dtype != np.float32:
            speech = speech.astype(np.float32)

        # data: (Nsamples,) -> (1, Nsamples)
        speech = speech[np.newaxis, :]

        # b. Forward Encoder
        enc, self.streaming_states = self.encoder(
            speech=speech,
            speech_length=np.array([speech.shape[1]]),
            states=self.streaming_states,
        )
        assert len(enc) == 1, len(enc)
        self.enc_feats += enc[0].tolist()
        best_hyps = self.beam_search(
            np.array(self.enc_feats, dtype=np.float32))
        self.encoder.increment()

        if best_hyps == []:
            return []
        else:
            return self._get_result(best_hyps[0])

    def simulate(self, speech: np.ndarray, print_every_hypo: bool = False):
        # This function will simulate streaming asr with the given audio.
        self.start()
        process_num = len(speech) // self.hop_size + 1
        logging.info(f'Processing audio with {process_num} processes.')
        padded_speech = self.pad(speech, length=process_num * self.hop_size)
        for i in range(process_num):
            start = self.hop_size * i
            end = self.hop_size * (i + 1)
            nbest = self(padded_speech[start:end])
            if print_every_hypo and nbest != []:
                logging.info(f'Result at position {i} : {nbest[0][0]}')

        nbest = self.end()
        return nbest

    def start(self):
        self.beam_search.__class__ = BatchBeamSearchOnlineSim
        self.encoder.reset()
        self.streaming_states = self.encoder.init_state()
        self.beam_search.start()

    def end(self):
        # compute final encoder process
        enc, *_ = self.encoder.forward_final(self.streaming_states)
        self.enc_feats += enc[0].tolist()
        best_hyps = self.batch_beam_search(
            np.array(self.enc_feats, dtype=np.float32))

        # initialize beam_search related parameters
        self.enc_feats = []
        self.streaming_states = {
            'buffer_before_downsampling': None,
            'buffer_after_downsampling': None,
            'prev_addin': None,
            'past_encoder_ctx': None,
        }
        self.beam_search.end()

        if best_hyps == []:
            return []
        else:
            return self._get_result(best_hyps[0])

    def _get_result(self, hyp: Hypothesis):
        token_int = hyp.yseq[2:-1].astype(np.int64).tolist()
        # remove blank symbol id, which is assumed to be 0
        token_int = list(filter(lambda x: x != 0, token_int))
        # Change integer-ids to tokens
        token = self.converter.ids2tokens(token_int)
        if self.tokenizer is not None:
            text = self.tokenizer.tokens2text(token)
        else:
            text = None
        results = [(text, token, token_int, hyp)]
        return results

    def pad(self, x, length=None):
        if length:
            base = np.zeros((length,))
        else:
            base = np.zeros((self.hop_size,))
        base[:len(x)] = x
        return base



from typing import (
    Optional,
    Tuple,
    List,
    Dict,
)

import onnxruntime
import numpy as np

from espnet_onnx.asr.frontend.frontend import Frontend
from espnet_onnx.asr.frontend.global_mvn import GlobalMVN
from espnet_onnx.asr.frontend.utterance_mvn import UtteranceMVN
from espnet_onnx.asr.scorer.interface import BatchScorerInterface
from espnet_onnx.utils.function import make_pad_mask
from espnet_onnx.utils.config import Config


class Tacotron2:
    def __init__(
        self,
        config: Config,
        providers: List[str],
        use_quantized: bool = False,
    ):
        self.config = config
        # load encoder and decoder.
        self.encoder = self._load_model(config.encoder, providers, use_quantized)
        self.predecoder = self._load_model(config.decoder.predecoder,
                                providers, use_quantized)
        self.decoder = self._load_model(config.decoder, providers, use_quantized)
        
        if self.config.decoder.postdecoder.onnx_export:
            self.postdecoder = self._load_model(config.decoder.postdecoder,
                                providers, use_quantized)
        else:
            self.postdecoder = None
        
        # HP
        self.input_names = [d.name for d in self.encoder.get_inputs()]
        self.use_sids = 'sids' in self.input_names
        self.use_lids = 'lids' in self.input_names
        self.use_feats = 'feats' in self.input_names
        self.dlayers = self.config.decoder.dlayers
        self.dunits = self.config.decoder.dunits
        self.threshold = self.config.decoder.threshold
        self.decoder_input_names = [d.name for d in self.decoder.get_inputs()]
        self.decoder_output_names = [d.name for d in self.decoder.get_outputs()]
    
    def _load_model(self, config, providers, use_quantized):
        if use_quantized:
            return onnxruntime.InferenceSession(
                config.quantized_model_path,
                providers=providers
            )
        else:
            return onnxruntime.InferenceSession(
                config.model_path,
                providers=providers
            )

    def __call__(
        self,
        text: np.ndarray,
        feats: np.ndarray = None,
        sids: np.ndarray = None,
        spembs:  np.ndarray = None,
        lids:  np.ndarray = None
    ):
        # compute encoder and initialize states
        input_enc = self.get_input_enc(text, feats, sids, spembs, lids)
        h = self.encoder.run(['h'], input_enc)[0]
        hs = h[None, :]
        
        idx = 0
        outs = []
        probs = []
        att_ws = []
        maxlen = int(h.shape[0] * self.config.decoder.maxlenratio)
        minlen = int(h.shape[0] * self.config.decoder.minlenratio)
        c_list, z_list, prev_out = self.init_state(hs)
        
        # compute decoder
        prev_att_w = None
        while True:
            idx += self.config.decoder.reduction_factor
            input_dec = self.get_input_dec(c_list, z_list, prev_att_w, prev_out)
            out, prob, a_prev, prev_out, *cz_states = \
                self.decoder.run(self.decoder_output_names, input_dec)
            
            if self.config.decoder.cumulate_att_w and prev_att_w is not None:
                prev_att_w = prev_att_w + a_prev
            else:
                prev_att_w = a_prev
            
            c_list, z_list = self._split(cz_states)
            
            outs += [out]
            probs += [prob]
            att_ws += [a_prev]
            
            # check whether to finish generation
            if int(sum(probs[-1] >= self.threshold)) > 0 or idx >= maxlen:
                # check mininum length
                if idx < minlen:
                    continue
                outs = np.concatenate(outs, axis=2)  # (1, odim, L)
                if self.postdecoder is not None:
                    outs = self.postdecoder.run(['out'], {
                        'x': outs
                    })[0] # (1, odim, L)
                probs = np.concatenate(probs, axis=0)
                att_ws = np.concatenate(att_ws, axis=0)
                break
        return dict(feat_gen=outs, prob=probs, att_w=att_ws)

    def get_input_enc(self, text, feats, sids, spembs, lids):
        ret = {'text': text }
        ret = self._set_input_dict(ret, 'feats', feats)
        ret = self._set_input_dict(ret, 'sids', sids)
        ret = self._set_input_dict(ret, 'spembs', spembs)
        ret = self._set_input_dict(ret, 'lids', lids)
        return ret

    def get_input_dec(self, c_list, z_list, a_prev, prev_in):
        ret = {}
        ret.update({
            f'c_prev_{i}': cl
            for i, cl in enumerate(c_list)
        })
        ret.update({
            f'z_prev_{i}': zl
            for i, zl in enumerate(z_list)
        })
        ret.update({
            'pceh': self.pre_compute_enc_h,
            'enc_h': self.enc_h,
            'mask': self.mask,
            'prev_in': prev_in
        })
        if a_prev is not None:
            ret.update(a_prev=a_prev)
        else:
            ret.update(a_prev=self.get_att_prev(self.enc_h))
        return ret

    def _set_input_dict(self, dic, key, value):
        if key in self.input_names:
            assert value is not None
            dic[key] = value
        return dic
    
    def zero_state(self):
        return np.zeros((1, self.dunits), dtype=np.float32)

    def get_att_prev(self, x, att_type=None):
        # x : (1, T, D)
        att_prev = 1.0 - make_pad_mask([x.shape[1]])
        att_prev = (
            att_prev / np.array([x.shape[1]])[..., None]).astype(np.float32)
        if att_type == 'location2d':
            att_prev = att_prev[..., None].reshape(-1, self.config.att_win, -1)
        if att_type in ('coverage', 'coverage_location'):
            att_prev = att_prev[:, None, :]
        return att_prev
    
    def init_state(self, x):
        # to support mutiple encoder asr mode, in single encoder mode,
        # convert torch.Tensor to List of torch.Tensor
        c_list = [self.zero_state()]
        z_list = [self.zero_state()]
        for _ in range(1, self.dlayers):
            c_list.append(self.zero_state())
            z_list.append(self.zero_state())

        prev_out = np.zeros((1, self.config.decoder.odim), dtype=np.float32)
        
        # compute predecoder
        self.pre_compute_enc_h = self.predecoder.run(
            ['pre_compute_enc_h'], { 'enc_h': x }
        )[0]
        self.enc_h = x
        self.mask = np.where(make_pad_mask(
                [x[0].shape[0]]) == 1, -float('inf'), 0).astype(np.float32)
        return c_list[:], z_list[:], prev_out

    def _split(self, status_lists):
        len_list = len(status_lists)
        c_list = status_lists[ : len_list // 2]
        z_list = status_lists[len_list // 2 : ]
        return c_list, z_list



import os
import logging

from espnet_onnx.utils.abs_model import AbsModel
from espnet_onnx.asr.model.encoder import get_encoder
from espnet_onnx.asr.model.decoder import get_decoder
from espnet_onnx.asr.model.lm import get_lm
from espnet_onnx.asr.model.joint_network import JointNetwork
from espnet_onnx.asr.scorer.ctc_prefix_scorer import CTCPrefixScorer
from espnet_onnx.asr.scorer.length_bonus import LengthBonus
from espnet_onnx.asr.scorer.interface import BatchScorerInterface
from espnet_onnx.asr.beam_search.beam_search import BeamSearch
from espnet_onnx.asr.beam_search.batch_beam_search import BatchBeamSearch
from espnet_onnx.asr.beam_search.beam_search_transducer import BeamSearchTransducer


class AbsASRModel(AbsModel): 
    def _build_beam_search(self, scorers, weights):
        if self.config.transducer.use_transducer_decoder:
            self.beam_search = BeamSearchTransducer(
                self.config.beam_search,
                self.config.token,
                scorers=scorers,
                weights=weights
            )
        else:
            self.beam_search = BeamSearch(
                self.config.beam_search,
                self.config.token,
                scorers=scorers,
                weights=weights,
            )
            non_batch = [
                k for k, v in self.beam_search.full_scorers.items()
                if not isinstance(v, BatchScorerInterface)
            ]
            if len(non_batch) == 0:
                self.beam_search.__class__ = BatchBeamSearch
                logging.info("BatchBeamSearch implementation is selected.")
            else:
                logging.warning(
                    f"As non-batch scorers {non_batch} are found, "
                    f"fall back to non-batch implementation."
                )
    
    def _build_model(self, providers, use_quantized):
        self.encoder = get_encoder(self.config.encoder, providers, use_quantized)
        decoder = get_decoder(self.config.decoder, providers, use_quantized)
        scorers = {'decoder': decoder}
        weights = {}
        if not self.config.transducer.use_transducer_decoder:
            ctc = CTCPrefixScorer(self.config.ctc, self.config.token.eos, providers, use_quantized)
            scorers.update(
                ctc=ctc,
                length_bonus=LengthBonus(len(self.config.token.list))
            )
            weights.update(
                decoder=self.config.weights.decoder,
                ctc=self.config.weights.ctc,
                length_bonus=self.config.weights.length_bonus,
            )
        else:
            joint_network = JointNetwork(self.config.joint_network, providers, use_quantized)
            scorers.update(joint_network=joint_network)
            
        lm = get_lm(self.config, providers, use_quantized)
        if lm is not None:
            scorers.update(lm=lm)
            weights.update(lm=self.config.weights.lm)

        self._build_beam_search(scorers, weights)
        self._build_tokenizer()
        self._build_token_converter()
        self.scorers = scorers
        self.weights = weights




from typing import (
    Optional,
    Tuple
)

import math
import torch
import torch.nn as nn
import torch.nn.functional as F

from espnet_onnx.utils.abs_model import AbsExportModel


class OnnxStyleMelGANVocoder(nn.Module, AbsExportModel):
    def __init__(
        self,
        model,
        use_z=False,
        **kwargs
    ):
        super().__init__()
        self.model = model
        self.in_channels = model.in_channels
        self.aux_channels = model.blocks[0].tade1.aux_conv[0].in_channels
        self.model_name = 'StyleMelGANVocoder'

    def forward(self, c: torch.Tensor) -> torch.Tensor:
        c = c.transpose(1, 0).unsqueeze(0)

        # prepare noise input
        noise_size = (
            1,
            self.in_channels,
            math.ceil(c.size(2) / self.model.noise_upsample_factor),
        )
        noise = torch.randn(*noise_size, dtype=torch.float)
        x = self.model.noise_upsample(noise)

        # NOTE(kan-bayashi): To remove pop noise at the end of audio, perform padding
        #    for feature sequence and after generation cut the generated audio. This
        #    requires additional computation but it can prevent pop noise.
        total_length = c.size(2) * self.model.upsample_factor
        c = F.pad(c, (0, x.size(2) - c.size(2)), "replicate")

        # This version causes pop noise.
        # x = x[:, :, :c.size(2)]

        for block in self.model.blocks:
            x, c = block(x, c)
        x = self.model.output_conv(x)[..., :total_length]

        return x.squeeze(0).transpose(1, 0)

    def get_dummy_inputs(self):
        c = torch.randn(100, self.aux_channels)
        return (c,)

    def get_input_names(self):
        return ['c']

    def get_output_names(self):
        return ['wav']

    def get_dynamic_axes(self):
        return {
            'c': {0: 'c_length'},
        }

    def get_model_config(self, path):
        return {
            'vocoder_type': 'StyleMelGANVocoder',
            'model_path': str(path / f'{self.model_name}.onnx')
        }



import os

from espnet2.text.char_tokenizer import CharTokenizer
from espnet2.text.phoneme_tokenizer import PhonemeTokenizer
from espnet2.text.sentencepiece_tokenizer import SentencepiecesTokenizer
from espnet2.text.word_tokenizer import WordTokenizer


def get_ngram_config(model):
    return {
        "use_ngram": True
    }

def get_weights_transducer(model):
    return {
        "lm": model.lm_weight
    }

def get_beam_config(model, minlenratio, maxlenratio):
    return {
        "beam_size": model.beam_size,
        "pre_beam_ratio": model.pre_beam_size / model.beam_size,
        "pre_beam_score_key": model.pre_beam_score_key,
        "maxlenratio": maxlenratio,
        "minlenratio": minlenratio
    }

def get_trans_beam_config(model):
    # check search algorithm
    search_type = ""
    search_args = {}
    if model.beam_size <= 1:
        search_type = "greedy"
    elif hasattr(model, 'max_sym_exp'):
        search_type = "tsd"
        search_args['max_sym_exp'] = model.max_sym_exp
    elif hasattr(model, 'u_max'):
        search_type = "alsd"
        search_args['u_max'] = model.u_max
    elif hasattr(model, 'nstep'):
        search_args['prefix_alpha'] = model.prefix_alpha
        if hasattr(model, 'expansion_gamma'):
            search_type = "maes"
            search_args['nstep'] = max(2, model.nstep)
            search_args['expansion_gamma'] = model.expansion_gamma
            search_args['expansion_beta'] = model.expansion_beta
        else:
            search_type = "nsc"
            search_args['nstep'] = model.nstep
    else:
        search_type = "default"

    return {
        "beam_size": model.beam_size,
        "search_type": search_type,
        "search_args": search_args,
        "score_norm": model.score_norm,
    }


def get_token_config(model):
    return {
        "sos": model.sos,
        "eos": model.eos,
        "blank": model.blank_id,
        "list": model.token_list
    }


def get_tokenizer_config(model, path):
    if model is None:
        return {}
    elif isinstance(model, SentencepiecesTokenizer):
        model_name = os.path.basename(model.model)
        return {
            "token_type": "bpe",
            "bpemodel": str(path.parent / model_name)
        }
    elif isinstance(model, WordTokenizer):
        return {
            "token_type": "word"
        }
    elif isinstance(model, CharTokenizer):
        return {
            "token_type": "char"
        }
    elif isinstance(model, PhonemeTokenizer):
        return {
            "token_type": "phn"
        }



from typing import Tuple
from typeguard import check_argument_types

import numpy as np

from espnet_onnx.utils.function import (
    make_pad_mask,
    mask_fill
)
from espnet_onnx.utils.config import Config

class UtteranceMVN:
    def __init__(
        self,
        config: Config
    ):
        assert check_argument_types()
        self.norm_means = config.norm_means
        self.norm_vars = config.norm_vars
        self.eps = config.eps

    def extra_repr(self):
        return f"norm_means={self.norm_means}, norm_vars={self.norm_vars}"

    def __call__(
        self, x: np.ndarray, ilens: np.ndarray = None
    ) -> Tuple[np.ndarray, np.ndarray]:
        """Forward function

        Args:
            x: (B, L, ...)
            ilens: (B,)

        """
        return utterance_mvn(
            x,
            ilens,
            norm_means=self.norm_means,
            norm_vars=self.norm_vars,
            eps=self.eps,
        )


def utterance_mvn(
    x: np.ndarray,
    ilens: np.ndarray = None,
    norm_means: bool = True,
    norm_vars: bool = False,
    eps: float = 1.0e-20,
) -> Tuple[np.ndarray, np.ndarray]:
    """Apply utterance mean and variance normalization

    Args:
        x: (B, T, D), assumed zero padded
        ilens: (B,)
        norm_means:
        norm_vars:
        eps:

    """
    if ilens is None:
        ilens = np.full([x.shape[0]], x.shape[1])
    ilens_ = ilens.reshape(-1, *[1 for _ in range(len(x.shape) - 1)])
    # Zero padding
    x = mask_fill(x, make_pad_mask(ilens, x, 1), 0.0)
    # mean: (B, 1, D)
    mean = x.sum(axis=1, keepdims=True) / ilens_

    if norm_means:
        x -= mean

        if norm_vars:
            var = np.power(x, 2).sum(axis=1, keepdims=True) / ilens_
            std = np.clip(np.sqrt(var), eps, None)
            x = x / np.sqrt(std)
        return x, ilens
    else:
        if norm_vars:
            y = x - mean
            y = mask_fill(y, make_pad_mask(ilens, y, 1), 0.0)
            var = np.power(y, 2).sum(axis=1, keepdims=True) / ilens_
            std = np.clip(np.sqrt(var), eps, None)
            x /= std
        return x, ilens



import os

import numpy as np
import torch
import torch.nn as nn
import torch.nn.functional as F

from espnet.nets.pytorch_backend.rnn.attentions import NoAtt

from espnet_onnx.utils.function import make_pad_mask
from espnet_onnx.export.layers.attention import get_attention, OnnxNoAtt, require_tanh
from espnet_onnx.export.layers.predecoder import PreDecoder
from espnet_onnx.utils.abs_model import AbsExportModel


def _apply_attention_constraint(
    e, last_attended_idx, backward_window=1, forward_window=3
):
    """Apply monotonic attention constraint.
    **Note** This function is copied from espnet.nets.pytorch_backend.rnn.attention.py
    """
    if e.size(0) != 1:
        raise NotImplementedError(
            "Batch attention constraining is not yet supported.")
    backward_idx = last_attended_idx - backward_window
    forward_idx = last_attended_idx + forward_window
    if backward_idx > 0:
        e[:, :backward_idx] = -float("inf")
    if forward_idx < e.size(1):
        e[:, forward_idx:] = -float("inf")
    return e


class RNNDecoder(nn.Module, AbsExportModel):
    def __init__(self, model, **kwargs):
        super().__init__()
        self.embed = model.embed
        self.submodel = []
        self.model = model
        self.num_encs = model.num_encs
        self.decoder_length = len(model.decoder)
        self.model_name = 'rnn_decoder'

        self.att_list = nn.ModuleList()
        for i, a in enumerate(model.att_list):
            self.att_list.append(get_attention(a))
            self.submodel.append(
                PreDecoder(a, idx=i)
            )

    def forward(self, vy, z_prev, c_prev, a_prev, pre_compute_enc_h, enc_h, mask):
        ey = self.embed(vy)  # utt list (1) x zdim
        if self.num_encs == 1:
            att_c, att_w = self.att_list[0](
                z_prev[0],
                a_prev[0],
                pre_compute_enc_h[0],
                enc_h[0],
                mask[0]
            )
        else:
            att_w = []
            for idx in range(self.num_encs):
                _att_c_list, _att_w = self.att_list[idx](
                    z_prev[0],
                    a_prev[idx],
                    pre_compute_enc_h[idx],
                    enc_h[idx],
                    mask[idx]
                )
                att_w.append(_att_w)

            att_c, _att_w = self.att_list[self.num_encs](
                z_prev[0],
                a_prev[self.num_encs],
                pre_compute_enc_h[self.num_encs],
                enc_h[self.num_encs],
                mask[self.num_encs]
            )
            att_w.append(_att_w)
        ey = torch.cat((ey, att_c), dim=1)  # utt(1) x (zdim + hdim)
        z_list, c_list = self.rnn_forward(
            ey, z_prev, c_prev
        )
        if self.model.context_residual:
            logits = self.model.output(
                torch.cat((z_list[-1], att_c), dim=-1)
            )
        else:
            logits = self.model.output(z_list[-1])
        logp = F.log_softmax(logits, dim=1).squeeze(0)
        return (
            logp,
            c_list,
            z_list,
            att_w,
        )

    def rnn_forward(self, ey, z_prev, c_prev):
        ret_z_list = []
        ret_c_list = []
        if self.model.dtype == "lstm":
            _z_list, _c_list = self.model.decoder[0](
                ey, (z_prev[0], c_prev[0]))
            ret_z_list.append(_z_list)
            ret_c_list.append(_c_list)
            for i in range(1, self.model.dlayers):
                _z_list, _c_list = self.model.decoder[i](
                    _z_list,
                    (z_prev[i], c_prev[i]),
                )
                ret_z_list.append(_z_list)
                ret_c_list.append(_c_list)
        else:
            _z_list = self.model.decoder[0](ey, z_prev[0])
            ret_z_list.append(_z_list)
            for i in range(1, self.model.dlayers):
                _z_list = self.model.decoder[i](
                    _z_list, z_prev[i]
                )
                ret_z_list.append(_z_list)
        return ret_z_list, ret_c_list
    
    def get_a_prev(self, feat_length, att):
        ret = torch.randn(1, feat_length)
        # if att.att_type == 'location2d':
        #     ret = torch.randn(1, att.att_win, feat_length)
        if att.att_type in ('coverage', 'coverage_location'):
            ret = torch.randn(1, 1, feat_length)
        return ret

    def get_dummy_inputs(self, enc_size):
        feat_length = 50
        vy = torch.LongTensor([1])
        z_prev = [torch.randn(1, self.model.dunits)
                  for _ in range(self.decoder_length)]
        a_prev = [self.get_a_prev(feat_length, att) for att in self.att_list]
        c_prev = [torch.randn(1, self.model.dunits)
                  for _ in range(self.decoder_length)]
        pre_compute_enc_h = [
            self.get_precompute_enc_h(i, feat_length)
            for i in range(self.num_encs)
        ]
        enc_h = [torch.randn(1, feat_length, enc_size)
                 for _ in range(self.num_encs)]
        _m = torch.from_numpy(np.where(make_pad_mask(
            [feat_length]) == 1, -float('inf'), 0)).type(torch.float32)
        mask = [_m for _ in range(self.num_encs)]
        return (
            vy, z_prev, c_prev,
            a_prev, pre_compute_enc_h,
            enc_h, mask
        )
    
    def get_precompute_enc_h(self, idx, feat_length):
        if isinstance(self.model.att_list[idx], NoAtt):
            return torch.randn(1, 1, 1)
        else:
            return torch.randn(1, feat_length, self.model.att_list[idx].mlp_enc.out_features)

    def get_input_names(self):
        ret = ['vy']
        ret += ['z_prev_%d' % i for i in range(self.decoder_length)]
        ret += ['c_prev_%d' % i for i in range(self.decoder_length)]
        ret += ['a_prev_%d' % i for i in range(self.num_encs)]
        ret += ['pceh_%d' % i for i in range(self.num_encs)]
        ret += ['enc_h_%d' % i for i in range(self.num_encs)]
        ret += ['mask_%d' % i for i in range(self.num_encs)]
        return ret

    def get_output_names(self):
        ret = ['logp']
        ret += ['c_list_%d' % i for i in range(self.decoder_length)]
        ret += ['z_list_%d' % i for i in range(self.decoder_length)]
        if self.num_encs == 1:
            ret += ['att_w']
        else:
            ret += ['att_w_%d' % i for i in range(self.num_encs + 1)]
        return ret

    def get_dynamic_axes(self):
        # input
        ret = {}
        ret.update({
            'a_prev_%d' % i: {
                a.get_dynamic_axes(): 'a_prev_%d_length' % i,
            }
            for i,a in enumerate(self.att_list)
        })
        ret.update({
            'pceh_%d' % d: {
                1: 'pceh_%d_length' % d,
            }
            for d in range(self.num_encs)
        })
        ret.update({
            'enc_h_%d' % d: {
                1: 'enc_h_%d_length' % d,
            }
            for d in range(self.num_encs)
        })
        ret.update({
            'mask_%d' % d: {
                0: 'mask_%d_length' % d,
                1: 'mask_%d_height' % d
            }
            for d in range(self.num_encs)
        })
        # output
        ret.update({
            'att_w_%d' % d: {
                1: 'att_w_%d_length' % d
            }
            for d in range(self.num_encs)
        })
        return ret

    def get_model_config(self, path):
        ret = {
            "dec_type": "RNNDecoder",
            "model_path": os.path.join(path, f'{self.model_name}.onnx'),
            "dlayers": self.model.dlayers,
            "odim": self.model.odim,
            "dunits": self.model.dunits,
            "decoder_length": self.decoder_length,
            "rnn_type": self.model.dtype,
            "predecoder": [
                {
                    "model_path": os.path.join(path, f'predecoder_{i}.onnx'),
                    "att_type": a.att_type
                }
                for i,a in enumerate(self.att_list)
                if not isinstance(a, OnnxNoAtt)
            ]
        }
        if hasattr(self.model, 'att_win'):
            ret.update(att_win=self.model.att_win)
        return ret



from setuptools import find_packages, setup


requirements = {
    "install": [
        "setuptools>=38.5.1",
        "librosa>=0.8.0",
        "onnxruntime>=1.4.0",
        "sentencepiece>=0.1.91,!=0.1.92",
        "typeguard>=2.7.0",
        "PyYAML>=5.1.2",
        "g2p-en",
        "jamo==0.4.1",  # For kss
        "numpy<=1.21.3",
        "espnet_tts_frontend" # required for TTS preprocess.
    ],
    "test": [
        "torch>=1.11.0",
        "espnet",
        "pytest"
    ]
}

setup(
    name="espnet_onnx",
    version="0.1.7",
    url="https://github.com/Masao-Someki/espnet_onnx",
    author="Masao Someki",
    author_email="masao.someki@gmail.com",
    description="ONNX Wrapper for ESPnet",
    long_description=open("README.md", "r", encoding="utf-8").read(),
    long_description_content_type="text/markdown",
    keywords="espnet onnxruntime",
    license="MIT",
    packages=find_packages(include=["espnet_onnx*"]),
    install_requires=requirements["install"],
    tests_require=requirements["test"],
    python_requires=">=3.7.0",
    classifiers=[
        "Programming Language :: Python",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "Development Status :: 4 - Beta",
        "Intended Audience :: Developers",
        "Operating System :: OS Independent",
        "License :: OSI Approved :: MIT License",
        "Topic :: Software Development :: Libraries :: Python Modules",
    ],
)


from typing import Any
from typing import Tuple
from typing import List

import warnings

import numpy as np


class ScorerInterface:
    """Scorer interface for beam search.
    The scorer performs scoring of the all tokens in vocabulary.
    Examples:
        * Search heuristics
            * :class:`espnet.nets.scorers.length_bonus.LengthBonus`
        * Decoder networks of the sequence-to-sequence models
            * :class:`espnet.nets.pytorch_backend.nets.transformer.decoder.Decoder`
            * :class:`espnet.nets.pytorch_backend.nets.rnn.decoders.Decoder`
        * Neural language models
            * :class:`espnet.nets.pytorch_backend.lm.transformer.TransformerLM`
            * :class:`espnet.nets.pytorch_backend.lm.default.DefaultRNNLM`
            * :class:`espnet.nets.pytorch_backend.lm.seq_rnn.SequentialRNNLM`
    """

    def init_state(self, x: np.ndarray) -> Any:
        """Get an initial state for decoding (optional).
        Args:
            x (np.ndarray): The encoded feature tensor
        Returns: initial state
        """
        return None

    def select_state(self, state: Any, i: int, new_id: int = None) -> Any:
        """Select state with relative ids in the main beam search.
        Args:
            state: Decoder state for prefix tokens
            i (int): Index to select a state in the main beam search
            new_id (int): New label index to select a state if necessary
        Returns:
            state: pruned state
        """
        return None if state is None else state[i]

    def score(
        self, y: np.ndarray, state: Any, x: np.ndarray
    ) -> Tuple[np.ndarray, Any]:
        """Score new token (required).
        Args:
            y (np.ndarray): 1D torch.int64 prefix tokens.
            state: Scorer state for prefix tokens
            x (np.ndarray): The encoder feature that generates ys.
        Returns:
            tuple[np.ndarray, Any]: Tuple of
                scores for next token that has a shape of `(n_vocab)`
                and next state for ys
        """
        raise NotImplementedError

    def final_score(self, state: Any) -> float:
        """Score eos (optional).
        Args:
            state: Scorer state for prefix tokens
        Returns:
            float: final score
        """
        return 0.0


class BatchScorerInterface(ScorerInterface):
    """Batch scorer interface."""

    def batch_init_state(self, x: np.ndarray) -> Any:
        """Get an initial state for decoding (optional).
        Args:
            x (np.ndarray): The encoded feature tensor
        Returns: initial state
        """
        return self.init_state(x)

    def batch_score(
        self, ys: np.ndarray, states: List[Any], xs: np.ndarray
    ) -> Tuple[np.ndarray, List[Any]]:
        """Score new token batch (required).
        Args:
            ys (np.ndarray): torch.int64 prefix tokens (n_batch, ylen).
            states (List[Any]): Scorer states for prefix tokens.
            xs (np.ndarray):
                The encoder feature that generates ys (n_batch, xlen, n_feat).
        Returns:
            tuple[np.ndarray, List[Any]]: Tuple of
                batchfied scores for next token with shape of `(n_batch, n_vocab)`
                and next state list for ys.
        """
        warnings.warn(
            "{} batch score is implemented through for loop not parallelized".format(
                self.__class__.__name__
            )
        )
        scores = list()
        outstates = list()
        for i, (y, state, x) in enumerate(zip(ys, states, xs)):
            score, outstate = self.score(y, state, x)
            outstates.append(outstate)
            scores.append(score)
        scores = np.concatenate(scores, axis=0).reshape(ys.shape[0], -1)
        return scores, outstates


class PartialScorerInterface(ScorerInterface):
    """Partial scorer interface for beam search.
    The partial scorer performs scoring when non-partial scorer finished scoring,
    and receives pre-pruned next tokens to score because it is too heavy to score
    all the tokens.
    Examples:
         * Prefix search for connectionist-temporal-classification models
             * :class:`espnet.nets.scorers.ctc.CTCPrefixScorer`
    """

    def score_partial(
        self, y: np.ndarray, next_tokens: np.ndarray, state: Any, x: np.ndarray
    ) -> Tuple[np.ndarray, Any]:
        """Score new token (required).
        Args:
            y (np.ndarray): 1D prefix token
            next_tokens (np.ndarray): torch.int64 next token to score
            state: decoder state for prefix tokens
            x (np.ndarray): The encoder feature that generates ys
        Returns:
            tuple[np.ndarray, Any]:
                Tuple of a score tensor for y that has a shape `(len(next_tokens),)`
                and next state for ys
        """
        raise NotImplementedError


class BatchPartialScorerInterface(BatchScorerInterface, PartialScorerInterface):
    """Batch partial scorer interface for beam search."""

    def batch_score_partial(
        self,
        ys: np.ndarray,
        next_tokens: np.ndarray,
        states: List[Any],
        xs: np.ndarray,
    ) -> Tuple[np.ndarray, Any]:
        """Score new token (required).
        Args:
            ys (np.ndarray): torch.int64 prefix tokens (n_batch, ylen).
            next_tokens (np.ndarray): torch.int64 tokens to score (n_batch, n_token).
            states (List[Any]): Scorer states for prefix tokens.
            xs (np.ndarray):
                The encoder feature that generates ys (n_batch, xlen, n_feat).
        Returns:
            tuple[np.ndarray, Any]:
                Tuple of a score tensor for ys that has a shape `(n_batch, n_vocab)`
                and next states for ys
        """
        raise NotImplementedError



import numpy as np
from .text_cleaner import TextCleaner


class CommonPreprocessor:
    def __init__(
        self,
        tokenizer,
        token_id_converter,
        cleaner_config,
    ):
        if cleaner_config.cleaner_types is not None:
            self.text_cleaner = TextCleaner(cleaner_config.cleaner_types)
        else:
            self.text_cleaner = None
        self.tokenizer = tokenizer
        self.token_id_converter = token_id_converter

    def __call__(self, text: str) -> np.ndarray:
        if self.text_cleaner is not None:
            text = self.text_cleaner(text)
            
        tokens = self.tokenizer.text2tokens(text)
        text_ints = self.token_id_converter.tokens2ids(tokens)
        return np.array(text_ints, dtype=np.int64)



from typing import (
    Optional,
    Tuple,
    List,
    Dict,
)

import onnxruntime
import numpy as np

from espnet_onnx.asr.frontend.frontend import Frontend
from espnet_onnx.asr.frontend.global_mvn import GlobalMVN
from espnet_onnx.asr.frontend.utterance_mvn import UtteranceMVN
from espnet_onnx.asr.scorer.interface import BatchScorerInterface
from espnet_onnx.utils.function import (
    subsequent_mask,
    make_pad_mask,
    mask_fill
)
from espnet_onnx.utils.config import Config


class VITS:
    def __init__(
        self,
        config: Config,
        providers: List[str],
        use_quantized: bool = False,
    ):
        self.config = config
        if use_quantized:
            self.model = onnxruntime.InferenceSession(
                self.config.quantized_model_path,
                providers=providers
            )
        else:
            self.model = onnxruntime.InferenceSession(
                self.config.model_path,
                providers=providers
            )

        self.input_names = [d.name for d in self.model.get_inputs()]
        self.use_sids = 'sids' in self.input_names
        self.use_lids = 'lids' in self.input_names
        self.use_feats = 'feats' in self.input_names

    def __call__(
        self,
        text: np.ndarray,
        feats: np.ndarray = None,
        sids: np.ndarray = None,
        spembs:  np.ndarray = None,
        lids:  np.ndarray = None
    ):
        output_names = ['wav', 'att_w', 'dur']
        input_dict = self.get_input_dict(
            text, feats, sids, spembs, lids)
        wav, att_w, dur = self.model.run(output_names, input_dict)
        return dict(wav=wav, att_w=att_w, dur=dur)

    def get_input_dict(self, text, feats, sids, spembs, lids):
        ret = {'text': text, 'text_length': np.array(
            [len(text)], dtype=np.int64)}
        ret = self._set_input_dict(ret, 'feats', feats)
        ret = self._set_input_dict(ret, 'sids', sids)
        ret = self._set_input_dict(ret, 'spembs', spembs)
        ret = self._set_input_dict(ret, 'lids', lids)
        return ret

    def _set_input_dict(self, dic, key, value):
        if key in self.input_names:
            assert value is not None
            dic[key] = value
        return dic



from .ctc import CTC
from .joint_network import JointNetwork

# encoder
from espnet2.asr.encoder.rnn_encoder import RNNEncoder as espnetRNNEncoder
from espnet2.asr.encoder.vgg_rnn_encoder import VGGRNNEncoder as espnetVGGRNNEncoder
from espnet2.asr.encoder.contextual_block_transformer_encoder import ContextualBlockTransformerEncoder as espnetContextualTransformer
from espnet2.asr.encoder.contextual_block_conformer_encoder import ContextualBlockConformerEncoder as espnetContextualConformer
from .encoders.rnn import RNNEncoder
from .encoders.xformer import XformerEncoder
from .encoders.contextual_block_xformer import ContextualBlockXformerEncoder

# decoder
from espnet2.asr.decoder.rnn_decoder import RNNDecoder as espnetRNNDecoder
from espnet2.asr.decoder.transformer_decoder import TransformerDecoder as espnetTransformerDecoder
from espnet2.asr.transducer.transducer_decoder import TransducerDecoder as espnetTransducerDecoder
from .decoders.rnn import (
    RNNDecoder,
    PreDecoder
)
from .decoders.xformer import XformerDecoder
from .decoders.transducer import TransducerDecoder

# lm
from espnet2.lm.seq_rnn_lm import SequentialRNNLM as espnetSequentialRNNLM
from espnet2.lm.transformer_lm import TransformerLM as espnetTransformerLM
from .language_models.seq_rnn import SequentialRNNLM
from .language_models.transformer import TransformerLM


def get_encoder(model, export_config):
    if isinstance(model, espnetRNNEncoder) or isinstance(model, espnetVGGRNNEncoder):
        return RNNEncoder(model, **export_config)
    elif isinstance(model, espnetContextualTransformer) or isinstance(model, espnetContextualConformer):
        return ContextualBlockXformerEncoder(model, **export_config)
    else:
        return XformerEncoder(model, **export_config)


def get_decoder(model, export_config):
    if isinstance(model, espnetRNNDecoder):
        return RNNDecoder(model, **export_config)
    elif isinstance(model, espnetTransducerDecoder):
        return TransducerDecoder(model, **export_config)
    else:
        return XformerDecoder(model, **export_config)


def get_lm(model, export_config):
    if isinstance(model, espnetSequentialRNNLM):
        return SequentialRNNLM(model, **export_config)
    elif isinstance(model, espnetTransformerLM):
        return TransformerLM(model, **export_config)



import numpy as np
import six

def subsequent_mask(size):
    """Create mask for subsequent steps (size, size).
    Modified from the original mask function to apply for fix-length mask.
    
    Args:
        size(int) : size of mask
    
    Examples:
        >>> subsequent_mask(3)
        [[1, 0, 0],
        [1, 1, 0],
        [1, 1, 1]]
    """
    return np.tril(np.ones((size, size)))


def mask_fill(arr, mask, mask_value):
    """Numpy implementation of torch.Tensor.masked_fill

    Args:
        arr (np.ndarray): Array to mask.
        mask (np.ndarray): Every value with 1 (or True) will mask with mask_value
        mask_value (number): Value to apply for mask

    Returns:
        np.ndarray: Masked array
    """
    arr[mask.astype(np.bool) == True] = mask_value
    return arr


def make_pad_mask(lengths, xs=None, dim=-1, xs_shape=None):
    """Make mask tensor containing indices of padded part.
    
    Args:
        lengths (np.ndarray or List): Batch of lengths (B,).
        xs (np.ndarray, optional): The reference tensor.
            If set, masks will be the same shape as this tensor.
        dim (int, optional): Dimension indicator of the above tensor.
            See the example.

    Returns:
        np.ndarray: Mask tensor containing indices of padded part.

    Examples:
        With only lengths.
        >>> lengths = [5, 3, 2]
        >>> make_pad_mask(lengths)
        masks = [[0, 0, 0, 0 ,0],
                 [0, 0, 0, 1, 1],
                 [0, 0, 1, 1, 1]]
                 
        With the reference tensor.
        >>> xs = np.zeros((3, 2, 4))
        >>> make_pad_mask(lengths, xs)
        array([[[0, 0, 0, 0],
                 [0, 0, 0, 0]],
                [[0, 0, 0, 1],
                 [0, 0, 0, 1]],
                [[0, 0, 1, 1],
                 [0, 0, 1, 1]]])
                 
        >>> xs = np.zeros((3, 2, 6))
        >>> make_pad_mask(lengths, xs)
        array([[[0, 0, 0, 0, 0, 1],
                 [0, 0, 0, 0, 0, 1]],
                [[0, 0, 0, 1, 1, 1],
                 [0, 0, 0, 1, 1, 1]],
                [[0, 0, 1, 1, 1, 1],
                 [0, 0, 1, 1, 1, 1]]])
                 
        With the reference tensor and dimension indicator.
        >>> xs = np.zeros((3, 6, 6))
        >>> make_pad_mask(lengths, xs, 1)
        array([[[0, 0, 0, 0, 0, 0],
                 [0, 0, 0, 0, 0, 0],
                 [0, 0, 0, 0, 0, 0],
                 [0, 0, 0, 0, 0, 0],
                 [0, 0, 0, 0, 0, 0],
                 [1, 1, 1, 1, 1, 1]],
                [[0, 0, 0, 0, 0, 0],
                 [0, 0, 0, 0, 0, 0],
                 [0, 0, 0, 0, 0, 0],
                 [1, 1, 1, 1, 1, 1],
                 [1, 1, 1, 1, 1, 1],
                 [1, 1, 1, 1, 1, 1]],
                [[0, 0, 0, 0, 0, 0],
                 [0, 0, 0, 0, 0, 0],
                 [1, 1, 1, 1, 1, 1],
                 [1, 1, 1, 1, 1, 1],
                 [1, 1, 1, 1, 1, 1],
                 [1, 1, 1, 1, 1, 1]]])
                 
        >>> make_pad_mask(lengths, xs, 2)
        array([[[0, 0, 0, 0, 0, 1],
                 [0, 0, 0, 0, 0, 1],
                 [0, 0, 0, 0, 0, 1],
                 [0, 0, 0, 0, 0, 1],
                 [0, 0, 0, 0, 0, 1],
                 [0, 0, 0, 0, 0, 1]],
                [[0, 0, 0, 1, 1, 1],
                 [0, 0, 0, 1, 1, 1],
                 [0, 0, 0, 1, 1, 1],
                 [0, 0, 0, 1, 1, 1],
                 [0, 0, 0, 1, 1, 1],
                 [0, 0, 0, 1, 1, 1]],
                [[0, 0, 1, 1, 1, 1],
                 [0, 0, 1, 1, 1, 1],
                 [0, 0, 1, 1, 1, 1],
                 [0, 0, 1, 1, 1, 1],
                 [0, 0, 1, 1, 1, 1],
                 [0, 0, 1, 1, 1, 1]]])
    """
    if xs is not None:
        base = np.zeros(xs.shape)
    else:
        base = np.zeros((len(lengths), max(lengths)))

    if len(base.shape) == 3 and dim == 1:
        base = base.transpose(0, 2, 1)

    for i in range(len(base)):
        base[i][..., lengths[i]:] = 1

    if len(base.shape) == 3 and dim == 1:
        base = base.transpose(0, 2, 1)

    return base


def topk(x: np.ndarray, k: int, require_value: bool = False):
    """Get indicies of topk.
    
    Args:
        x (np.ndarray)
        k (int)
        
    Examples:
        >>> a = np.array([3,6,1,7])
        >>> topk(a, 2)
        array([3, 1])
        >>> b = np.array([[3,6,2,7],
                          [6,2,4,8],
                          [1,1,7,3]])
        >>> topk(b, 2)
        array([[3,1],
               [3,0],
               [2,3]])
    """
    topk_index = np.argpartition(x, x.shape[-1] - k, axis=-1)[..., -k:]
    if require_value:
        return np.take_along_axis(x, topk_index, axis=-1), topk_index
    else:
        return topk_index


def pad_sequence(yseqs, batch_first=False, padding_value=0):
    """Numpy implementation of torch.pad_sequence

    Args:
        yseqs (np.ndarray): List of array. (B, *)
        batch_first (bool):
        padding_value (int, optional): Padding value. Defaults to 0.

    Returns:
        np.ndarray

    Examples:
        >>> a = np.ones(25, 300)
        >>> b = np.ones(22, 300)
        >>> c = np.ones(15, 300)
        >>> pad_sequence([a, b, c]).size()
        (25, 3, 300)
        
        >>> pad_sequence([a, b, c], batch_first=True).size()
        (3, 25, 300)
    """
    if len(yseqs) == 1:
        return np.array(yseqs)

    max_idx = np.argmax([y.shape[0] for y in yseqs])
    max_shape = yseqs[max_idx].shape
    base = np.ones((len(yseqs), *max_shape)) * padding_value
    for i, y in enumerate(yseqs):
        base[i][:y.shape[0]] = y
    if batch_first:
        return base
    else:
        return base.transpose(1, 0, *np.arange(2, len(base.shape)))


def is_prefix(x, pref) -> bool:
    """Check if pref is a prefix of x.
    
    Args:
        x: Label ID sequence.
        pref: Prefix label ID sequence.
        
    Returns:
        : Whether pref is a prefix of x.
    """
    if len(pref) >= len(x):
        return False

    for i in range(len(pref)):
        if pref[i] != x[i]:
            return False

    return True


def recombine_hyps(hyps):
    """Recombine hypotheses with same label ID sequence.
    
    Args:
        hyps: Hypotheses.
        
    Returns:
       final: Recombined hypotheses.
    """
    final = []

    for hyp in hyps:
        seq_final = [f.yseq for f in final if f.yseq]

        if hyp.yseq in seq_final:
            seq_pos = seq_final.index(hyp.yseq)

            final[seq_pos].score = np.logaddexp(
                final[seq_pos].score, hyp.score)
            
        else:
            final.append(hyp)

    return final


def select_k_expansions(
    hyps,
    logps,
    beam_size,
    gamma,
    beta,
):
    """Return K hypotheses candidates for expansion from a list of hypothesis.
    K candidates are selected according to the extended hypotheses probabilities
    and a prune-by-value method. Where K is equal to beam_size + beta.
    
    Args:
        hyps: Hypotheses.
        beam_logp: Log-probabilities for hypotheses expansions.
        beam_size: Beam size.
        gamma: Allowed logp difference for prune-by-value method.
        beta: Number of additional candidates to store.
        
    Return:
        k_expansions: Best K expansion hypotheses candidates.
    """
    k_expansions = []

    for i, hyp in enumerate(hyps):
        hyp_i = [(int(k), hyp.score + float(logp))
                 for k, logp in enumerate(logps[i])]
        k_best_exp = max(hyp_i, key=lambda x: x[1])[1]

        k_expansions.append(
            sorted(
                filter(lambda x: (k_best_exp - gamma) <= x[1], hyp_i),
                key=lambda x: x[1],
                reverse=True,
            )[: beam_size + beta]
        )

    return k_expansions


def subtract(x, subset):
    """Remove elements of subset if corresponding label ID sequence already exist in x.
    
    Args:
        x: Set of hypotheses.
        subset: Subset of x.
        
    Returns:
       final: New set of hypotheses.
    """
    final = []

    for x_ in x:
        if any(x_.yseq == sub.yseq for sub in subset):
            continue
        final.append(x_)

    return final

def narrow(arr: np.ndarray, axis: int, start: int, length: int):
    """Numpy implementation of torch.narrow

    Args:
        arr (np.ndarray): the array to narrow
        axis (int): the dimension along which to narrow
        start (int): the starting dimension
        length (int): the distance to the ending dimension
    """
    return arr.take(np.arange(start, start+length), axis=axis)

def end_detect(ended_hyps, i, M=3, D_end=np.log(1 * np.exp(-10))):
    """End detection.
    described in Eq. (50) of S. Watanabe et al
    "Hybrid CTC/Attention Architecture for End-to-End Speech Recognition"
    :param ended_hyps:
    :param i:
    :param M:
    :param D_end:
    :return:
    """
    if len(ended_hyps) == 0:
        return False
    count = 0
    best_hyp = sorted(ended_hyps, key=lambda x: x["score"], reverse=True)[0]
    for m in six.moves.range(M):
        # get ended_hyps with their length is i - m
        hyp_length = i - m
        hyps_same_length = [x for x in ended_hyps if len(x["yseq"]) == hyp_length]
        if len(hyps_same_length) > 0:
            best_hyp_same_length = sorted(
                hyps_same_length, key=lambda x: x["score"], reverse=True
            )[0]
            if best_hyp_same_length["score"] - best_hyp["score"] < D_end:
                count += 1

    if count == M:
        return True
    else:
        return False


#!/usr/bin/env python3

"""Griffin-Lim related modules."""

# Copyright 2019 Tomoki Hayashi
#  Apache 2.0  (http://www.apache.org/licenses/LICENSE-2.0)
# Modified by Masao Someki
# Copyright (c) 2022 Masao someki

import logging
from functools import partial
from typing import Optional

import librosa
import numpy as np
from typeguard import check_argument_types
from packaging.version import parse as V
from espnet_onnx.utils.config import Config

EPS = 1e-10


def logmel2linear(
    lmspc: np.ndarray,
    fs: int,
    n_fft: int,
    n_mels: int,
    fmin: int = None,
    fmax: int = None,
) -> np.ndarray:
    """Convert log Mel filterbank to linear spectrogram.
    Args:
        lmspc: Log Mel filterbank (T, n_mels).
        fs: Sampling frequency.
        n_fft: The number of FFT points.
        n_mels: The number of mel basis.
        f_min: Minimum frequency to analyze.
        f_max: Maximum frequency to analyze.
    Returns:
        Linear spectrogram (T, n_fft // 2 + 1).
    """
    assert lmspc.shape[1] == n_mels
    fmin = 0 if fmin is None else fmin
    fmax = fs / 2 if fmax is None else fmax
    mspc = np.power(10.0, lmspc)
    mel_basis = librosa.filters.mel(
        sr=fs, n_fft=n_fft, n_mels=n_mels, fmin=fmin, fmax=fmax
    )
    inv_mel_basis = np.linalg.pinv(mel_basis)
    return np.maximum(EPS, np.dot(inv_mel_basis, mspc.T).T)


def griffin_lim(
    spc: np.ndarray,
    n_fft: int,
    n_shift: int,
    win_length: int = None,
    window: Optional[str] = "hann",
    n_iter: Optional[int] = 32,
) -> np.ndarray:
    """Convert linear spectrogram into waveform using Griffin-Lim.
    Args:
        spc: Linear spectrogram (T, n_fft // 2 + 1).
        n_fft: The number of FFT points.
        n_shift: Shift size in points.
        win_length: Window length in points.
        window: Window function type.
        n_iter: The number of iterations.
    Returns:
        Reconstructed waveform (N,).
    """
    # assert the size of input linear spectrogram
    assert spc.shape[1] == n_fft // 2 + 1
    if V(librosa.__version__) >= V("0.7.0"):
        # use librosa's fast Grriffin-Lim algorithm
        spc = np.abs(spc.T)
        y = librosa.griffinlim(
            S=spc,
            n_iter=n_iter,
            hop_length=n_shift,
            win_length=win_length,
            window=window,
            center=True if spc.shape[1] > 1 else False,
        )
    else:
        # use slower version of Grriffin-Lim algorithm
        logging.warning(
            "librosa version is old. use slow version of Grriffin-Lim algorithm."
            "if you want to use fast Griffin-Lim, please update librosa via "
            "`source ./path.sh && pip install librosa==0.7.0`."
        )
        cspc = np.abs(spc).astype(np.complex).T
        angles = np.exp(2j * np.pi * np.random.rand(*cspc.shape))
        y = librosa.istft(cspc * angles, n_shift, win_length, window=window)
        for i in range(n_iter):
            angles = np.exp(
                1j
                * np.angle(librosa.stft(y, n_fft, n_shift, win_length, window=window))
            )
            y = librosa.istft(cspc * angles, n_shift, win_length, window=window)

    return y


class Spectrogram2Waveform(object):
    """Spectrogram to waveform conversion module."""
    def __init__(
        self,
        config: Config
    ):
        """Initialize module.
        Args:
            fs: Sampling frequency.
            n_fft: The number of FFT points.
            n_shift: Shift size in points.
            n_mels: The number of mel basis.
            win_length: Window length in points.
            window: Window function type.
            f_min: Minimum frequency to analyze.
            f_max: Maximum frequency to analyze.
            griffin_lim_iters: The number of iterations.
        """
        assert check_argument_types()
        
        self.fs = config.fs
        self.logmel2linear = (
            partial(
                logmel2linear, fs=config.fs, n_fft=config.n_fft, n_mels=config.n_mels, fmin=config.fmin, fmax=config.fmax
            )
            if config.n_mels is not None
            else None
        )
        self.griffin_lim = partial(
            griffin_lim,
            n_fft=config.n_fft,
            n_shift=config.n_shift,
            win_length=config.win_length,
            window=config.window,
            n_iter=config.n_iter,
        )
        self.params = dict(
            n_fft=config.n_fft,
            n_shift=config.n_shift,
            win_length=config.win_length,
            window=config.window,
            n_iter=config.n_iter,
        )
        if config.n_mels is not None:
            self.params.update(fs=config.fs, n_mels=config.n_mels, fmin=config.fmin, fmax=config.fmax)

    def __repr__(self):
        retval = f"{self.__class__.__name__}("
        for k, v in self.params.items():
            retval += f"{k}={v}, "
        retval += ")"
        return retval

    def __call__(self, spc: np.ndarray) -> np.ndarray:
        """Convert spectrogram to waveform.
        Args:
            spc: Log Mel filterbank (T_feats, n_mels)
                or linear spectrogram (T_feats, n_fft // 2 + 1).
        Returns:
            Tensor: Reconstructed waveform (T_wav,).
        """
        if self.logmel2linear is not None:
            spc = self.logmel2linear(spc)
        return self.griffin_lim(spc)



from abc import ABC

from typing import List

import os
import glob
import logging
import onnxruntime
import warnings

from espnet_onnx.utils.abs_model import AbsModel
from espnet_onnx.asr.frontend.global_mvn import GlobalMVN
from espnet_onnx.tts.model.preprocess.common_processor import CommonPreprocessor
from espnet_onnx.tts.model.duration_calculator import DurationCalculator
from espnet_onnx.tts.model.tts_model import get_tts_model
from espnet_onnx.tts.model.vocoders.vocoder import Vocoder
from espnet_onnx.tts.model.vocoders.griffin_lim import Spectrogram2Waveform
from espnet_onnx.utils.config import (
    get_config,
    get_tag_config
)
from espnet_onnx.asr.postprocess.build_tokenizer import build_tokenizer
from espnet_onnx.asr.postprocess.token_id_converter import TokenIDConverter


class AbsTTSModel(AbsModel):

    def _check_argument(self, tag_name, model_dir):
        self.model_dir = model_dir

        if tag_name is None and model_dir is None:
            raise ValueError('tag_name or model_dir should be defined.')

        if tag_name is not None:
            tag_config = get_tag_config()
            if tag_name not in tag_config.keys():
                raise RuntimeError(f'Model path for tag_name "{tag_name}" is not set on tag_config.yaml.'
                                   + 'You have to export to onnx format with `espnet_onnx.export.asr.export_asr.ModelExport`,'
                                   + 'or have to set exported model path in tag_config.yaml.')
            self.model_dir = tag_config[tag_name]

    def _load_config(self):
        config_file = glob.glob(os.path.join(self.model_dir, 'config.*'))[0]
        self.config = get_config(config_file)

    def _build_tokenizer(self):
        if self.config.tokenizer.token_type is None:
            self.tokenizer = None
        elif self.config.tokenizer.token_type == 'bpe':
            self.tokenizer = build_tokenizer(
                'bpe', self.config.tokenizer.bpemodel)
        else:
            self.tokenizer = build_tokenizer(
                **self.config.tokenizer.dic)

    def _build_token_converter(self):
        self.converter = TokenIDConverter(token_list=self.config.token.list)
    
    def _build_vocoder(self, providers, use_quantized):
        self.vocoder = None
        if self.config.vocoder.vocoder_type == 'Spectrogram2Waveform':
            self.vocoder = Spectrogram2Waveform(self.config.vocoder)
        elif self.config.vocoder.vocoder_type == 'PretrainedPWGVocoder':
            raise RuntimeError('Currently, PWGVocoder is not supported.')
        elif self.config.vocoder.vocoder_type == 'OnnxVocoder':
            self.vocoder = Vocoder(self.config.vocoder, providers, use_quantized)
        else:
            raise RuntimeError(f'vocoder type {self.config.vocoder_type} is not supported.')
    
    def _build_normalizer(self):
        self.normalize = None
        if self.config.normalize.use_normalize:
            self.normalize = GlobalMVN(self.config.normalize)

    def _build_model(self, providers, use_quantized):
        # build tts model such as vits
        self.tts_model = get_tts_model(
            self.config.tts_model, providers, use_quantized)

        self._build_tokenizer()
        self._build_token_converter()
        self.preprocess = CommonPreprocessor(
            tokenizer=self.tokenizer,
            token_id_converter=self.converter,
            cleaner_config=self.config.text_cleaner,
        )
        self.duration_calculator = DurationCalculator()
        self._build_normalizer()
        self._build_vocoder(providers, use_quantized)

    def _check_ort_version(self, providers: List[str]):
        # check cpu
        if onnxruntime.get_device() == 'CPU' and 'CPUExecutionProvider' not in providers:
            raise RuntimeError(
                'If you want to use GPU, then follow `How to use GPU on espnet_onnx` chapter in readme to install onnxruntime-gpu.')

        # check GPU
        if onnxruntime.get_device() == 'GPU' and providers == ['CPUExecutionProvider']:
            warnings.warn(
                'Inference will be executed on the CPU. Please provide gpu providers. Read `How to use GPU on espnet_onnx` in readme in detail.')

        logging.info(f'Providers [{" ,".join(providers)}] detected.')



import os
import six

import numpy as np
import torch
import torch.nn as nn
import torch.nn.functional as F

from espnet.nets.pytorch_backend.transformer.subsampling import Conv2dSubsampling
from espnet.nets.pytorch_backend.transformer.subsampling import Conv2dSubsampling2
from espnet.nets.pytorch_backend.transformer.subsampling import Conv2dSubsampling6
from espnet.nets.pytorch_backend.transformer.subsampling import Conv2dSubsampling8
from espnet.nets.pytorch_backend.rnn.encoders import (RNNP, RNN, VGG2L)
from espnet2.asr.encoder.rnn_encoder import RNNEncoder as espnetRNNEncoder
from espnet2.asr.encoder.vgg_rnn_encoder import VGGRNNEncoder as espnetVGGRNNEncoder
from espnet2.asr.frontend.default import DefaultFrontend
from espnet2.layers.global_mvn import GlobalMVN
from espnet2.layers.utterance_mvn import UtteranceMVN

from espnet_onnx.utils.abs_model import AbsExportModel


class OnnxRNNP(nn.Module):
    def __init__(self, model):
        super().__init__()
        self.model = model
        self.subsample = model.subsample
        is_bidirectional = 2 if model.bidir else 1
        self.initial_state = torch.zeros(is_bidirectional * 1, 1, model.cdim)

    def forward(self, xs_pad: torch.Tensor, ilens: torch.Tensor):
        """Inference version of RNNEncoder.
        Since batch_size is always 1 in inference with onnx, pad_sequence related
        functions can be removed.
        """
        elayer_states = []
        for layer in six.moves.range(self.model.elayers):
            rnn = getattr(
                self.model, ("birnn" if self.model.bidir else "rnn") + str(layer))
            ys, states = rnn(xs_pad, hx=self.initial_state)
            elayer_states.append(states)
            # ys: utt list of frame x cdim x 2 (2: means bidirectional)
            sub = self.subsample[layer + 1]
            if sub > 1:
                ys = ys[:, ::sub]
                ilens = torch.div(ilens, sub, rounding_mode='trunc')
            # (sum _utt frame_utt) x dim
            projection_layer = getattr(self.model, "bt%d" % layer)
            projected = projection_layer(ys.contiguous().view(-1, ys.size(2)))
            xs_pad = projected.view(ys.size(0), ys.size(1), -1)
            if layer < self.model.elayers - 1:
                xs_pad = torch.tanh(xs_pad)

        return xs_pad, ilens, elayer_states  # x: utt list of frame x dim


class OnnxRNN(nn.Module):
    def __init__(self, model):
        super().__init__()
        self.nbrnn = model.nbrnn
        self.l_last = model.l_last
        is_bidirectional = 2 if model.nbrnn.bidirectional else 1
        num_layers = model.nbrnn.num_layers
        self.initial_state = torch.zeros(
            is_bidirectional * num_layers,
            1,
            model.nbrnn.hidden_size
        )

    def forward(self, xs_pad: torch.Tensor, ilens: torch.Tensor):
        if not isinstance(ilens, torch.Tensor):
            ilens = torch.tensor(ilens)

        ys, states = self.nbrnn(xs_pad, hx=self.initial_state)
        # ys: utt list of frame x cdim x 2 (2: means bidirectional)
        # (sum _utt frame_utt) x dim
        projected = torch.tanh(
            self.l_last(ys.contiguous().view(-1, ys.size(2)))
        )
        xs_pad = projected.view(ys.size(0), ys.size(1), -1)
        return xs_pad, ilens, states  # x: utt list of frame x dim


class OnnxVGG2l(nn.Module):
    def __init__(self, model):
        super().__init__()
        self.model = model
        self.in_channel = model.in_channel
    
    def forward(self, xs_pad, ilens, **kwargs):
        if not isinstance(ilens, torch.Tensor):
            ilens = torch.tensor(ilens)
        # x: utt x frame x dim
        # x: utt x 1 (input channel num) x frame x dim
        xs_pad = xs_pad.view(
            xs_pad.size(0),
            xs_pad.size(1),
            self.in_channel,
            xs_pad.size(2) // self.in_channel,
        ).transpose(1, 2)
        # NOTE: max_pool1d ?
        xs_pad = F.relu(self.model.conv1_1(xs_pad))
        xs_pad = F.relu(self.model.conv1_2(xs_pad))
        xs_pad = F.max_pool2d(xs_pad, 2, stride=2, ceil_mode=True)
        xs_pad = F.relu(self.model.conv2_1(xs_pad))
        xs_pad = F.relu(self.model.conv2_2(xs_pad))
        xs_pad = F.max_pool2d(xs_pad, 2, stride=2, ceil_mode=True)
        # x: utt_list of frame (remove zeropaded frames) x (input channel num x dim)
        xs_pad = xs_pad.transpose(1, 2)
        xs_pad = xs_pad.contiguous().view(
            xs_pad.size(0), xs_pad.size(1), xs_pad.size(2) * xs_pad.size(3)
        )
        
        ilens = torch.ceil(ilens / 4).type(torch.long)
        return xs_pad, ilens, None  # no state in this layer


class RNNEncoderLayer(nn.Module):
    def __init__(self, layer):
        super().__init__()
        if isinstance(layer, RNNP):
            self.layer = OnnxRNNP(layer)
        elif isinstance(layer, RNN):
            self.layer = OnnxRNN(layer)
        elif isinstance(layer, VGG2L):
            self.layer = OnnxVGG2l(layer)

    def forward(self, *args, **kwargs):
        return self.layer(*args, **kwargs)


class RNNEncoder(nn.Module, AbsExportModel):
    def __init__(self, model, feats_dim=80, **kwargs):
        super().__init__()
        self.model = model
        self.model_name = 'rnn_encoder'
        self.enc = nn.ModuleList()
        self.feats_dim = feats_dim
        for e in model.enc:
            self.enc.append(RNNEncoderLayer(e))

    def forward(self, feats, ilens):
        current_states = []
        for module in self.enc:
            feats, ilens, states = module(feats, ilens)
            current_states.append(states)
        return feats, ilens, current_states

    def get_output_size(self):
        return self.model._output_size

    def get_dummy_inputs(self):
        feats = torch.randn(1, 100, self.feats_dim)
        feats_length = torch.LongTensor([feats.size(1)])
        return (feats, feats_length)

    def get_input_names(self):
        return ['feats', 'feats_length']

    def get_output_names(self):
        return ['encoder_out', 'encoder_out_lens']

    def get_dynamic_axes(self):
        return {
            'feats': {
                1: 'feats_length'
            }
        }

    def get_model_config(self, asr_model=None, path=None):
        ret = {}
        ret.update(
            enc_type='RNNEncoder',
            model_path=os.path.join(path, f'{self.model_name}.onnx'),
            is_vggrnn=isinstance(self.model, espnetVGGRNNEncoder),
            frontend=self.get_frontend_config(asr_model.frontend),
            do_normalize=asr_model.normalize is not None,
            do_preencoder=asr_model.preencoder is not None,
            do_postencoder=asr_model.postencoder is not None
        )
        if ret['do_normalize']:
            ret.update(normalize=self.get_norm_config(
                asr_model.normalize, path))
        # Currently preencoder, postencoder is not supported.
        # if ret['do_preencoder']:
        #     ret.update(preencoder=get_preenc_config(self.model.preencoder))
        # if ret['do_postencoder']:
        #     ret.update(postencoder=get_postenc_config(self.model.postencoder))
        return ret

    def get_frontend_config(self, frontend):
        # currently only default config is supported.
        assert isinstance(
            frontend, DefaultFrontend), 'Currently only DefaultFrontend is supported.'

        stft_config = dict(
            n_fft=frontend.stft.n_fft,
            win_length=frontend.stft.win_length,
            hop_length=frontend.stft.hop_length,
            window=frontend.stft.window,
            center=frontend.stft.center,
            onesided=frontend.stft.onesided,
            normalized=frontend.stft.normalized,
        )
        logmel_config = frontend.logmel.mel_options
        logmel_config.update(log_base=frontend.logmel.log_base)
        return {
            "stft": stft_config,
            "logmel": logmel_config
        }

    def get_norm_config(self, normalize, path):
        if isinstance(normalize, GlobalMVN):
            return {
                "type": "gmvn",
                "norm_means": normalize.norm_means,
                "norm_vars": normalize.norm_vars,
                "eps": normalize.eps,
                "stats_file": str(path.parent / 'feats_stats.npz')
            }
        elif isinstance(normalize, UtteranceMVN):
            return {
                "type": "utterance_mvn",
                "norm_means": normalize.norm_means,
                "norm_vars": normalize.norm_vars,
                "eps": normalize.eps,
            }



from typing import List
from typing import Any
from typing import Tuple
from typing import Optional
from typing import Dict
from typing import Union

import numpy as np
import onnxruntime

from espnet_onnx.asr.scorer.interface import BatchScorerInterface
from espnet_onnx.asr.beam_search.hyps import (
    Hypothesis,
    TransducerHypothesis,
    ExtendedHypothesis
)
from espnet_onnx.utils.function import subsequent_mask
from espnet_onnx.utils.config import Config


class TransducerDecoder(BatchScorerInterface):
    def __init__(
        self,
        config: Config,
        providers: List[str],
        use_quantized: bool = False
    ):
        """Onnx support for espnet2.asr.decoder.transformer_decoder

        Args:
            config (Config):
            use_quantized (bool): Flag to use quantized model
        """
        if use_quantized:
            self.decoder = onnxruntime.InferenceSession(
                config.quantized_model_path,
                providers=providers
            )
        else:
            self.decoder = onnxruntime.InferenceSession(
                config.model_path,
                providers=providers
            )
        self.n_layers = config.n_layers
        self.odim = config.odim
        self.dtype = config.dtype
        self.output_names = ['sequence'] \
            + sorted([d.name for d in self.decoder.get_outputs() if 'h_cache' in d.name]) \
            + sorted([d.name for d in self.decoder.get_outputs()
                      if 'c_cache' in d.name])

    def score(
        self, hyp: Hypothesis, cache: Dict[str, Any]
    ) -> Tuple[np.ndarray, Tuple[np.ndarray, Optional[np.ndarray]], np.ndarray]:
        """One-step forward hypothesis.

        Args:
            hyp: Hypothesis.
            cache: Pairs of (dec_out, state) for each label sequence. (key)

        Returns:
            dec_out: Decoder output sequence. (1, D_dec)
            new_state: Decoder hidden states. ((N, 1, D_dec), (N, 1, D_dec))
            label: Label ID for LM. (1,)

        """
        label = np.full((1, 1), hyp.yseq[-1], dtype=np.int64)
        str_labels = "_".join(list(map(str, hyp.yseq)))

        if str_labels in cache:
            dec_out, dec_state = cache[str_labels]
        else:
            input_dict = self.get_input_dict(label, hyp.dec_state)
            dec_out, * \
                next_states = self.decoder.run(self.output_names, input_dict)
            dec_state = self.split(next_states)
            cache[str_labels] = (dec_out, dec_state)

        return dec_out[0][0], dec_state, label[0]

    def batch_score(
        self,
        hyps: Union[List[Hypothesis], List[ExtendedHypothesis]],
        dec_states: Tuple[np.ndarray, Optional[np.ndarray]],
        cache: Dict[str, Any],
        use_lm: bool,
    ) -> Tuple[np.ndarray, Tuple[np.ndarray, np.ndarray], np.ndarray]:
        """One-step forward hypotheses.

        Args:
            hyps: Hypotheses.
            states: Decoder hidden states. ((N, B, D_dec), (N, B, D_dec))
            cache: Pairs of (dec_out, dec_states) for each label sequences. (keys)
            use_lm: Whether to compute label ID sequences for LM.

        Returns:
            dec_out: Decoder output sequences. (B, D_dec)
            dec_states: Decoder hidden states. ((N, B, D_dec), (N, B, D_dec))
            lm_labels: Label ID sequences for LM. (B,)

        """
        final_batch = len(hyps)

        process = []
        done = [None] * final_batch

        for i, hyp in enumerate(hyps):
            str_labels = "_".join(list(map(str, hyp.yseq)))

            if str_labels in cache:
                done[i] = cache[str_labels]
            else:
                process.append((str_labels, hyp.yseq[-1], hyp.dec_state))

        if process:
            labels = np.array([[p[1]] for p in process], dtype=np.int64)
            p_dec_states = self.create_batch_states(
                self.init_state(labels.shape[0]), [p[2] for p in process]
            )
            input_dict = self.get_input_dict(labels, p_dec_states)
            dec_out, * \
                next_states = self.decoder.run(self.output_names, input_dict)
            new_states = self.split(next_states)

        j = 0
        for i in range(final_batch):
            if done[i] is None:
                state = self.select_state(new_states, j)
                done[i] = (dec_out[j], state)
                cache[process[j][0]] = (dec_out[j], state)
                j += 1

        dec_out = np.concatenate([d[0] for d in done], axis=0)
        dec_states = self.create_batch_states(dec_states, [d[1] for d in done])
        if use_lm:
            lm_labels = np.array(
                [h.yseq[-1] for h in hyps], dtype=np.int64
            ).view(final_batch, 1)

            return dec_out, dec_states, lm_labels

        return dec_out, dec_states, None

    def get_input_dict(self, labels, states):
        ret = {
            'labels': labels,
            'h_cache': states[0],
            'c_cache': states[1]
        }
        return ret

    def select_state(self, states: Tuple[np.ndarray, Optional[np.ndarray]], idx: int) -> Tuple[np.ndarray, Optional[np.ndarray]]:
        """Get specified ID state from decoder hidden states.

        Args:
            states: Decoder hidden states. ((N, B, D_dec), (N, B, D_dec))
            idx: State ID to extract.

        Returns:
            : Decoder hidden state for given ID.
              ((N, 1, D_dec), (N, 1, D_dec))

        """
        return (
            states[0][:, idx: idx + 1, :],
            states[1][:, idx: idx + 1, :] if self.dtype == "lstm" else None,
        )

    def init_state(self, batch_size: int = 1) -> Tuple[np.ndarray, Optional[np.ndarray]]:
        """Initialize decoder states.

        Args:
            batch_size: Batch size.

        Returns:
            : Initial decoder hidden states. ((N, B, D_dec), (N, B, D_dec))

        """
        h_n = np.zeros((self.n_layers, batch_size,
                        self.odim), dtype=np.float32)

        if self.dtype == "lstm":
            c_n = np.zeros((self.n_layers, batch_size,
                            self.odim), dtype=np.float32)
            return (h_n, c_n)

        return (h_n, None)

    def create_batch_states(
        self,
        states: Tuple[np.ndarray, Optional[np.ndarray]],
        new_states: List[Tuple[np.ndarray, Optional[np.ndarray]]],
        check_list: Optional[List] = None,
    ) -> Tuple[np.ndarray, Optional[np.ndarray]]:
        """Create decoder hidden states.

        Args:
            states: Decoder hidden states. ((N, B, D_dec), (N, B, D_dec))
            new_states: Decoder hidden states. [N x ((1, D_dec), (1, D_dec))]

        Returns:
            states: Decoder hidden states. ((N, B, D_dec), (N, B, D_dec))

        """
        return (
            np.concatenate([s[0] for s in new_states], axis=1),
            np.concatenate([s[1] for s in new_states], axis=1)
            if self.dtype == "lstm"
            else None,
        )

    def split(self, states):
        if self.dtype == "lstm":
            h_next = states[0]
            c_next = states[1]
            return (h_next, c_next)
        else:
            return (states, None)
